# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import
from __future__ import print_function

import os

import setuptools


def smart_find_packages(package_list):
    """Run find_packages the way we intend."""
    packages = []
    for pkg in package_list.strip().split("\n"):
        pkg_path = pkg.replace('.', os.path.sep)
        packages.append(pkg)
        packages.extend(
            ['%s.%s' % (pkg, f) for f in setuptools.find_packages(pkg_path)]
        )
    return "\n".join(set(packages))



# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import
from __future__ import print_function

import os

import fixtures

from pbr.hooks import files
from pbr.tests import base
from pbr.tests import fixtures as pbr_fixtures


class FilesConfigTest(base.BaseTestCase):

    def setUp(self):
        super(FilesConfigTest, self).setUp()

        pkg_fixture = fixtures.PythonPackage(
            "fake_package",
            [
                ("fake_module.py", b""),
                ("other_fake_module.py", b""),
            ],
        )
        self.useFixture(pkg_fixture)
        pkg_etc = os.path.join(pkg_fixture.base, 'etc')
        pkg_ansible = os.path.join(
            pkg_fixture.base, 'ansible', 'kolla-ansible', 'test'
        )
        dir_spcs = os.path.join(pkg_fixture.base, 'dir with space')
        dir_subdir_spc = os.path.join(
            pkg_fixture.base, 'multi space', 'more spaces'
        )
        pkg_sub = os.path.join(pkg_etc, 'sub')
        subpackage = os.path.join(
            pkg_fixture.base, 'fake_package', 'subpackage'
        )
        os.makedirs(pkg_sub)
        os.makedirs(subpackage)
        os.makedirs(pkg_ansible)
        os.makedirs(dir_spcs)
        os.makedirs(dir_subdir_spc)
        with open(os.path.join(pkg_etc, "foo"), 'w') as foo_file:
            foo_file.write("Foo Data")
        with open(os.path.join(pkg_sub, "bar"), 'w') as foo_file:
            foo_file.write("Bar Data")
        with open(os.path.join(pkg_ansible, "baz"), 'w') as baz_file:
            baz_file.write("Baz Data")
        with open(os.path.join(subpackage, "__init__.py"), 'w') as foo_file:
            foo_file.write("# empty")
        with open(os.path.join(dir_spcs, "file with spc"), 'w') as spc_file:
            spc_file.write("# empty")
        with open(os.path.join(dir_subdir_spc, "file with spc"), 'w') as file_:
            file_.write("# empty")

        self.useFixture(pbr_fixtures.Chdir(pkg_fixture.base))

    def test_implicit_auto_package(self):
        config = {'files': {}}
        files.FilesConfig(config, 'fake_package').run()
        self.assertIn('subpackage', config['files']['packages'])

    def test_auto_package(self):
        config = {'files': {'packages': 'fake_package'}}
        files.FilesConfig(config, 'fake_package').run()
        self.assertIn('subpackage', config['files']['packages'])

    def test_data_files_globbing(self):
        config = {'files': {'data_files': '\n  etc/pbr = etc/*'}}
        files.FilesConfig(config, 'fake_package').run()
        self.assertIn(
            "\n'etc/pbr/' = \n 'etc/foo'\n'etc/pbr/sub' = \n 'etc/sub/bar'",
            config['files']['data_files'],
        )

    def test_data_files_with_spaces(self):
        config = {
            'files': {'data_files': "\n  'i like spaces' = 'dir with space'/*"}
        }
        files.FilesConfig(config, 'fake_package').run()
        self.assertIn(
            "\n'i like spaces/' = \n 'dir with space/file with spc'",
            config['files']['data_files'],
        )

    def test_data_files_with_spaces_subdirectories(self):
        # test that we can handle whitespace in subdirectories
        data_files = "\n 'one space/two space' = 'multi space/more spaces'/*"
        expected = (
            "\n'one space/two space/' = "
            "\n 'multi space/more spaces/file with spc'"
        )
        config = {'files': {'data_files': data_files}}
        files.FilesConfig(config, 'fake_package').run()
        self.assertIn(expected, config['files']['data_files'])

    def test_data_files_with_spaces_quoted_components(self):
        # test that we can quote individual path components
        data_files = (
            "\n'one space'/'two space' = 'multi space'/'more spaces'/*"
        )
        expected = (
            "\n'one space/two space/' = "
            "\n 'multi space/more spaces/file with spc'"
        )
        config = {'files': {'data_files': data_files}}
        files.FilesConfig(config, 'fake_package').run()
        self.assertIn(expected, config['files']['data_files'])

    def test_data_files_globbing_source_prefix_in_directory_name(self):
        # We want to test that the string, "docs", is not replaced in a
        # subdirectory name, "sub-docs"
        config = {'files': {'data_files': "\n  share/ansible = ansible/*"}}
        files.FilesConfig(config, 'fake_package').run()
        self.assertIn(
            "\n'share/ansible/' = "
            "\n'share/ansible/kolla-ansible' = "
            "\n'share/ansible/kolla-ansible/test' = "
            "\n 'ansible/kolla-ansible/test/baz'",
            config['files']['data_files'],
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Utilities to paste over differences between Python versions."""

from __future__ import absolute_import
from __future__ import print_function

import re

_packaging_lib = None

PACKAGING_LIB_PACKAGING = 'packaging'
PACKAGING_LIB_LEGACY = 'pkg_resources'


def _get_packaging_lib():
    global _packaging_lib

    if _packaging_lib is not None:
        return _packaging_lib

    # packaging should almost always be available since setuptools vendors it
    # and has done so since forever
    #
    # https://github.com/pypa/setuptools/commit/84c9006110e53c84296a05741edb7b9edd305f12
    try:
        import packaging  # noqa

        _packaging_lib = PACKAGING_LIB_PACKAGING
        return _packaging_lib
    except ImportError:
        pass

    # pkg_resources is our fallback. This will always be available on older
    # Python versions since it's part of setuptools.
    try:
        import pkg_resources  # noqa

        _packaging_lib = PACKAGING_LIB_LEGACY
        return _packaging_lib
    except ImportError:
        pass

    raise RuntimeError(
        'Failed to find a library for parsing packaging information. This '
        'should not happen. Please report a bug against pbr.'
    )


def extract_project_name(requirement_line):
    packaging_lib = _get_packaging_lib()
    if packaging_lib == PACKAGING_LIB_PACKAGING:
        import packaging.requirements

        try:
            requirement = packaging.requirements.Requirement(requirement_line)
        except ValueError:
            return None

        # the .project_name attribute is not part of the
        # packaging.requirements.Requirement API so we mimic it
        #
        # https://github.com/pypa/setuptools/blob/v80.9.0/pkg_resources/__init__.py#L2918
        return re.sub('[^A-Za-z0-9.]+', '-', requirement.name)
    else:  # PACKAGING_LIB_LEGACY
        import pkg_resources

        try:
            requirement = pkg_resources.Requirement.parse(requirement_line)
        except ValueError:
            return None
        return requirement.project_name


def parse_version(version):
    packaging_lib = _get_packaging_lib()
    if packaging_lib == PACKAGING_LIB_PACKAGING:
        import packaging.version

        return packaging.version.Version(version)
    else:  # PACKAGING_LIB_LEGACY
        import pkg_resources

        return pkg_resources.parse_version(version)


def evaluate_marker(marker):
    packaging_lib = _get_packaging_lib()
    if packaging_lib == PACKAGING_LIB_PACKAGING:
        import packaging.markers

        try:
            return packaging.markers.Marker(marker).evaluate()
        except packaging.markers.InvalidMarker as e:
            # setuptools expects a SyntaxError here, so we do the same.
            # we can't chain the exceptions since that is a Python 3 only thing
            raise SyntaxError(e)
    else:  # PACKAGING_LIB_LEGACY
        import pkg_resources

        return pkg_resources.evaluate_marker(marker)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Most of this code is copied from setuptools ([1], [2]), licensed under the
# MIT license
#
# [1] https://github.com/pypa/setuptools/blob/v67.8.0/setuptools/command/easy_install.py
# [2] https://github.com/pypa/setuptools/blob/v67.8.0/setuptools/_distutils/spawn.py

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import re
import shlex
import subprocess
import sys
import textwrap
import warnings

import pbr._compat.metadata


shebang_pattern = re.compile('^#!.*python[0-9.]*([ \t].*)?$')
"""
Pattern matching a Python interpreter indicated in first line of a script.
"""


def isascii(s):
    try:
        s.encode('ascii')
    except UnicodeError:
        return False
    return True


def find_executable(executable, path=None):
    """Tries to find 'executable' in the directories listed in 'path'.

    A string listing directories separated by 'os.pathsep'; defaults to
    os.environ['PATH'].  Returns the complete filename or None if not found.
    """
    _, ext = os.path.splitext(executable)
    if (sys.platform == 'win32') and (ext != '.exe'):
        executable = executable + '.exe'

    if os.path.isfile(executable):
        return executable

    if path is None:
        path = os.environ.get('PATH', None)
        if path is None:
            try:
                path = os.confstr("CS_PATH")
            except (AttributeError, ValueError):
                # os.confstr() or CS_PATH is not available
                path = os.defpath
        # bpo-35755: Don't use os.defpath if the PATH environment variable is
        # set to an empty string

    # PATH='' doesn't match, whereas PATH=':' looks in the current directory
    if not path:
        return None

    paths = path.split(os.pathsep)
    for p in paths:
        f = os.path.join(p, executable)
        if os.path.isfile(f):
            # the file exists, we have a shot at spawn working
            return f
    return None


class CommandSpec(list):
    """
    A command spec for a #! header, specified as a list of arguments akin to
    those passed to Popen.
    """

    options = []  # type: list[str]
    split_args = dict()  # type: dict[str, bool]

    @classmethod
    def best(cls):
        """
        Choose the best CommandSpec class based on environmental conditions.
        """
        return cls

    @classmethod
    def _sys_executable(cls):
        _default = os.path.normpath(sys.executable)
        return os.environ.get('__PYVENV_LAUNCHER__', _default)

    @classmethod
    def from_param(cls, param):
        """
        Construct a CommandSpec from a parameter to build_scripts, which may
        be None.
        """
        if isinstance(param, cls):
            return param
        if isinstance(param, list):
            return cls(param)
        if param is None:
            return cls.from_environment()
        # otherwise, assume it's a string.
        return cls.from_string(param)

    @classmethod
    def from_environment(cls):
        return cls([cls._sys_executable()])

    @classmethod
    def from_string(cls, string):
        """
        Construct a command spec from a simple string representing a command
        line parseable by shlex.split.
        """
        items = shlex.split(string, **cls.split_args)
        return cls(items)

    def install_options(self, script_text):
        self.options = shlex.split(self._extract_options(script_text))
        cmdline = subprocess.list2cmdline(self)
        if not isascii(cmdline):
            self.options[:0] = ['-x']

    @staticmethod
    def _extract_options(orig_script):
        """
        Extract any options from the first line of the script.
        """
        first = (orig_script + '\n').splitlines()[0]
        match = shebang_pattern.match(first)
        options = match.group(1) or '' if match else ''
        return options.strip()

    def as_header(self):
        return self._render(self + list(self.options))

    @staticmethod
    def _strip_quotes(item):
        _QUOTES = '"\''
        for q in _QUOTES:
            if item.startswith(q) and item.endswith(q):
                return item[1:-1]
        return item

    @staticmethod
    def _render(items):
        cmdline = subprocess.list2cmdline(
            CommandSpec._strip_quotes(item.strip()) for item in items
        )
        return '#!' + cmdline + '\n'


sys_executable = CommandSpec._sys_executable


class WindowsCommandSpec(CommandSpec):
    split_args = dict(posix=False)


_wsgi_text = """#PBR Generated from %(group)r

import threading

from %(module_name)s import %(import_target)s

if __name__ == "__main__":
    import argparse
    import socket
    import sys
    import wsgiref.simple_server as wss

    parser = argparse.ArgumentParser(
        description=%(import_target)s.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        usage='%%(prog)s [-h] [--port PORT] [--host IP] -- [passed options]')
    parser.add_argument('--port', '-p', type=int, default=8000,
                        help='TCP port to listen on')
    parser.add_argument('--host', '-b', default='',
                        help='IP to bind the server to')
    parser.add_argument('args',
                        nargs=argparse.REMAINDER,
                        metavar='-- [passed options]',
                        help="'--' is the separator of the arguments used "
                        "to start the WSGI server and the arguments passed "
                        "to the WSGI application.")
    args = parser.parse_args()
    if args.args:
        if args.args[0] == '--':
            args.args.pop(0)
        else:
            parser.error("unrecognized arguments: %%s" %% ' '.join(args.args))
    sys.argv[1:] = args.args
    server = wss.make_server(args.host, args.port, %(invoke_target)s())

    print("*" * 80)
    print("STARTING test server %(module_name)s.%(invoke_target)s")
    url = "http://%%s:%%d/" %% (server.server_name, server.server_port)
    print("Available at %%s" %% url)
    print("DANGER! For testing only, do not use in production")
    print("*" * 80)
    sys.stdout.flush()

    server.serve_forever()
else:
    application = None
    app_lock = threading.Lock()

    with app_lock:
        if application is None:
            application = %(invoke_target)s()

"""

_script_text = """# PBR Generated from %(group)r

import sys

from %(module_name)s import %(import_target)s


if __name__ == "__main__":
    sys.exit(%(invoke_target)s())
"""

# the following allows us to specify different templates per entry
# point group when generating pbr scripts.
ENTRY_POINTS_MAP = {
    'console_scripts': _script_text,
    'gui_scripts': _script_text,
    'wsgi_scripts': _wsgi_text,
}


def generate_script(group, entry_point, header, template):
    """Generate the script based on the template.

    :param str group: The entry-point group name, e.g., "console_scripts".
    :param str header: The first line of the script, e.g.,
        "!#/usr/bin/env python".
    :param str template: The script template.
    :returns: The templated script content
    :rtype: str
    """
    if not entry_point.attrs or len(entry_point.attrs) > 2:
        raise ValueError(
            "Script targets must be of the form "
            "'func' or 'Class.class_method'."
        )

    script_text = template % {
        'group': group,
        'module_name': entry_point.module_name,
        'import_target': entry_point.attrs[0],
        'invoke_target': '.'.join(entry_point.attrs),
    }
    return header + script_text


class ScriptWriter:
    """
    Encapsulates behavior around writing entry point scripts for console and
    gui apps.
    """

    command_spec_class = CommandSpec

    @classmethod
    def get_script_args(cls, dist, executable=None, wininst=False):
        # NOTE(stephenfin): This was deprecated upstream. We opt not to
        # deprecate it here.
        writer = (WindowsScriptWriter if wininst else ScriptWriter).best()
        header = cls.get_script_header("", executable, wininst)
        return writer.get_args(dist, header)

    @classmethod
    def get_script_header(cls, script_text, executable=None, wininst=False):
        # NOTE(stephenfin): This was deprecated upstream. We opt not to
        # deprecate it here.
        if wininst:
            executable = "python.exe"
        return cls.get_header(script_text, executable)

    @classmethod
    def get_args(cls, dist, header=None):
        """
        Yield write_script() argument tuples for a distribution's
        console_scripts and gui_scripts entry points.
        """
        # NOTE(stephenfin): This is modified from upstream to add support for
        # wsgi-scripts. The Windows version is unchanged.
        if header is None:
            header = cls.get_header()

        for group, template in ENTRY_POINTS_MAP.items():
            for name, ep in pbr._compat.metadata.get_entry_points(dist, group):
                cls._ensure_safe_name(name)
                yield (name, generate_script(group, ep, header, template))

    @staticmethod
    def _ensure_safe_name(name):
        """
        Prevent paths in *_scripts entry point names.
        """
        has_path_sep = re.search(r'[\\/]', name)
        if has_path_sep:
            raise ValueError("Path separators not allowed in script names")

    @classmethod
    def best(cls):
        """
        Select the best ScriptWriter for this environment.
        """
        if sys.platform == 'win32' or (os.name == 'java' and os._name == 'nt'):
            return WindowsScriptWriter.best()
        else:
            return cls

    @classmethod
    def _get_script_args(cls, type_, name, header, script_text):
        # Simply write the stub with no extension.
        yield (name, header + script_text)

    @classmethod
    def get_header(cls, script_text="", executable=None):
        """Create a #! line, getting options (if any) from script_text"""
        cmd = cls.command_spec_class.best().from_param(executable)
        cmd.install_options(script_text)
        return cmd.as_header()


class WindowsScriptWriter(ScriptWriter):
    template = textwrap.dedent(
        r"""
        # EASY-INSTALL-ENTRY-SCRIPT: %(spec)r,%(group)r,%(name)r
        import re
        import sys

        # for compatibility with easy_install; see #2198
        __requires__ = %(spec)r

        try:
            from importlib.metadata import distribution
        except ImportError:
            try:
                from importlib_metadata import distribution
            except ImportError:
                from pkg_resources import load_entry_point


        def importlib_load_entry_point(spec, group, name):
            dist_name, _, _ = spec.partition('==')
            matches = (
                entry_point
                for entry_point in distribution(dist_name).entry_points
                if entry_point.group == group and entry_point.name == name
            )
            return next(matches).load()


        globals().setdefault('load_entry_point', importlib_load_entry_point)


        if __name__ == '__main__':
            sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
            sys.exit(load_entry_point(%(spec)r, %(group)r, %(name)r)())
        """
    ).lstrip()

    command_spec_class = WindowsCommandSpec

    @classmethod
    def get_args(cls, dist, header=None):
        """
        Yield write_script() argument tuples for a distribution's
        console_scripts and gui_scripts entry points.
        """
        if header is None:
            header = cls.get_header()
        spec = str(dist.as_requirement())
        for type_ in 'console', 'gui':
            group = type_ + '_scripts'
            for name, ep in pbr._compat.metadata.get_entry_points(dist, group):
                cls._ensure_safe_name(name)
                script_text = cls.template % {
                    'spec': spec,
                    'group': group,
                    'name': name,
                }
                args = cls._get_script_args(type_, name, header, script_text)
                for res in args:
                    yield res

    @classmethod
    def best(cls):
        """
        Select the best ScriptWriter suitable for Windows
        """
        # NOTE(stephenfin): We don't support the
        # WindowsExecutableLauncherWriter since it has a significant dependency
        # on pkg_resources
        return cls

    @classmethod
    def _get_script_args(cls, type_, name, header, script_text):
        "For Windows, add a .py extension"
        ext = dict(console='.pya', gui='.pyw')[type_]
        if ext not in os.environ['PATHEXT'].lower().split(';'):
            msg = (
                "{ext} not listed in PATHEXT; scripts will not be "
                "recognized as executables."
            ).format(ext=ext)
            warnings.warn(msg, UserWarning)
        old = ['.pya', '.py', '-script.py', '.pyc', '.pyo', '.pyw', '.exe']
        old.remove(ext)
        header = cls._adjust_header(type_, header)
        blockers = [name + x for x in old]
        yield name + ext, header + script_text, 't', blockers

    @classmethod
    def _adjust_header(cls, type_, orig_header):
        """
        Make sure 'pythonw' is used for gui and 'python' is used for
        console (regardless of what sys.executable is).
        """
        pattern = 'pythonw.exe'
        repl = 'python.exe'
        if type_ == 'gui':
            pattern, repl = repl, pattern
        pattern_ob = re.compile(re.escape(pattern), re.IGNORECASE)
        new_header = pattern_ob.sub(string=orig_header, repl=repl)
        return new_header if cls._use_header(new_header) else orig_header

    @staticmethod
    def _use_header(new_header):
        """
        Should _adjust_header use the replaced header?

        On non-windows systems, always use. On
        Windows systems, only use the replaced header if it resolves
        to an executable on the system.
        """
        clean_header = new_header[2:-1].strip('"')
        return sys.platform != 'win32' or find_executable(clean_header)


# for backward-compatibility
get_script_args = ScriptWriter.get_script_args
get_script_header = ScriptWriter.get_script_header



# Copyright (c) 2013 New Dream Network, LLC (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright (C) 2013 Association of Universities for Research in Astronomy
#                    (AURA)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided
#        with the distribution.
#
#     3. The name of AURA and its representatives may not be used to
#        endorse or promote products derived from this software without
#        specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS

import os

from pbr.tests.functional import base


class TestPbrJson(base.BaseWheelTestCase):

    def test_metadata_directory_has_pbr_json(self):
        # Build the path to the scripts directory
        pbr_json = os.path.join(
            self.extracted_wheel_dir, 'pbr_testpackage-0.0.dist-info/pbr.json'
        )
        self.assertTrue(os.path.exists(pbr_json))



# Copyright (c) 2013 New Dream Network, LLC (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright (C) 2013 Association of Universities for Research in Astronomy
#                    (AURA)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided
#        with the distribution.
#
#     3. The name of AURA and its representatives may not be used to
#        endorse or promote products derived from this software without
#        specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS

import os
import sysconfig

from pbr.tests.functional import base

try:
    import importlib.machinery

    get_suffixes = importlib.machinery.all_suffixes
# NOTE(JayF): ModuleNotFoundError only exists in Python 3.6+, not in 2.7
except ImportError:
    import imp

    # NOTE(JayF) imp.get_suffixes returns a list of three-tuples;
    # we need the first value from each tuple.

    def get_suffixes():
        return [x[0] for x in imp.get_suffixes]


def get_soabi():
    soabi = None
    try:
        soabi = sysconfig.get_config_var('SOABI')
        arch = sysconfig.get_config_var('MULTIARCH')
    except IOError:
        pass
    if soabi and arch and 'pypy' in sysconfig.get_scheme_names():
        soabi = '%s-%s' % (soabi, arch)
    if soabi is None and 'pypy' in sysconfig.get_scheme_names():
        # NOTE(sigmavirus24): PyPy only added support for the SOABI config var
        # to sysconfig in 2015. That was well after 2.2.1 was published in the
        # Ubuntu 14.04 archive.
        for suffix in get_suffixes():
            if suffix.startswith('.pypy') and suffix.endswith('.so'):
                soabi = suffix.split('.')[1]
                break
    return soabi


class TestCExtension(base.BaseWheelTestCase):

    def test_generates_c_extensions(self):
        built_package_dir = os.path.join(
            self.extracted_wheel_dir, 'pbr_testpackage'
        )
        static_object_filename = 'testext.so'
        ext_suffix = sysconfig.get_config_var('EXT_SUFFIX')
        if ext_suffix is not None:
            static_object_filename = 'testext' + ext_suffix
        else:
            soabi = get_soabi()
            if soabi:
                static_object_filename = 'testext.{0}.so'.format(soabi)
        static_object_path = os.path.join(
            built_package_dir, static_object_filename
        )

        self.assertTrue(os.path.exists(built_package_dir))
        self.assertTrue(os.path.exists(static_object_path))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import
from __future__ import print_function

import sys

from pbr import pbr_json
from pbr.tests import base

if sys.version_info >= (3, 3):
    from unittest import mock
else:
    import mock  # noqa


class TestJsonContent(base.BaseTestCase):
    @mock.patch('pbr.git._run_git_functions', return_value=True)
    @mock.patch('pbr.git.get_git_short_sha', return_value="123456")
    @mock.patch('pbr.git.get_is_release', return_value=True)
    def test_content(self, mock_get_is, mock_get_git, mock_run):
        cmd = mock.Mock()
        pbr_json.write_pbr_json(cmd, "basename", "pbr.json")
        cmd.write_file.assert_called_once_with(
            'pbr', 'pbr.json', '{"git_version": "123456", "is_release": true}'
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Metadata parsing."""

from __future__ import absolute_import
from __future__ import print_function

from collections import namedtuple
import json
import sys

_metadata_lib = None

METADATA_LIB_STDLIB = 'importlib.metadata'
METADATA_LIB_BACKPORT = 'importlib_metadata'
METADATA_LIB_LEGACY = 'pkg_resources'

entrypoint = namedtuple('entrypoint', ['module_name', 'attrs'])
dist = namedtuple('dist', ['egg_base', 'egg_info', 'egg_name', 'egg_version'])


def _get_metadata_lib():
    """Retrieve the correct metadata library to use."""
    global _metadata_lib

    if _metadata_lib is not None:
        return _metadata_lib

    # try importlib.metadata first. This will be available from the stdlib
    # starting in python >= 3.8
    if sys.version_info >= (3, 8):
        _metadata_lib = METADATA_LIB_STDLIB
        return _metadata_lib

    # try importlib_metadata next. This must be installed from PyPI and we
    # don't vendor it, but if available it will be preferred since later
    # versions of pkg_resources issue very annoying deprecation warnings
    try:
        import importlib_metadata  # noqa

        _metadata_lib = METADATA_LIB_BACKPORT
        return _metadata_lib
    except ImportError:
        pass

    # pkg_resources is our fallback. This will always be available on older
    # Python versions since it's part of setuptools.
    try:
        import pkg_resources  # noqa

        _metadata_lib = METADATA_LIB_LEGACY
        return _metadata_lib
    except ImportError:
        pass

    raise RuntimeError(
        'Failed to find a library for loading metadata. This should not '
        'happen. Please report a bug against pbr.'
    )


def get_distributions():
    metadata_lib = _get_metadata_lib()
    if metadata_lib == METADATA_LIB_STDLIB:
        import importlib.metadata

        data = sorted(
            importlib.metadata.distributions(),
            key=lambda x: x.metadata['name'].lower(),
        )
    elif metadata_lib == METADATA_LIB_BACKPORT:
        import importlib_metadata

        data = sorted(
            importlib_metadata.distributions(),
            key=lambda x: x.metadata['name'].lower(),
        )
    else:  # METADATA_LIB_LEGACY
        import pkg_resources

        data = sorted(
            pkg_resources.working_set,
            key=lambda dist: dist.project_name.lower(),
        )

    return list(data)


class PackageNotFound(Exception):
    def __init__(self, package_name):
        self.package_name = package_name

    def __str__(self):
        return 'Package {0} not installed'.format(self.package_name)


def get_metadata(package_name):
    metadata_lib = _get_metadata_lib()
    if metadata_lib == METADATA_LIB_STDLIB:
        import importlib.metadata

        try:
            data = importlib.metadata.distribution(package_name).metadata[
                'pbr.json'
            ]
        except importlib.metadata.PackageNotFoundError:
            raise PackageNotFound(package_name)
    elif metadata_lib == METADATA_LIB_BACKPORT:
        import importlib_metadata

        try:
            data = importlib_metadata.distribution(package_name).metadata[
                'pbr.json'
            ]
        except importlib_metadata.PackageNotFoundError:
            raise PackageNotFound(package_name)
    else:  # METADATA_LIB_LEGACY
        import pkg_resources

        try:
            data = pkg_resources.get_distribution(package_name).get_metadata(
                'pbr.json'
            )
        except pkg_resources.DistributionNotFound:
            raise PackageNotFound(package_name)

    try:
        return json.loads(data)
    except Exception:
        # TODO(stephenfin): We should log an error here. Can we still use
        # distutils.log in the future?
        return None


def get_version(package_name):
    metadata_lib = _get_metadata_lib()
    if metadata_lib == METADATA_LIB_STDLIB:
        import importlib.metadata

        try:
            return importlib.metadata.distribution(package_name).version
        except importlib.metadata.PackageNotFoundError:
            raise PackageNotFound(package_name)
    elif metadata_lib == METADATA_LIB_BACKPORT:
        import importlib_metadata

        try:
            return importlib_metadata.distribution(package_name).version
        except importlib_metadata.PackageNotFoundError:
            raise PackageNotFound(package_name)
    else:  # METADATA_LIB_LEGACY
        import pkg_resources

        try:
            return pkg_resources.get_distribution(package_name).version
        except pkg_resources.DistributionNotFound:
            raise PackageNotFound(package_name)


def get_entry_points(dist, group):
    metadata_lib = _get_metadata_lib()

    if metadata_lib == METADATA_LIB_STDLIB:
        import importlib.metadata

        try:
            dist = importlib.metadata.Distribution.at(dist.egg_info)
        except importlib.metadata.PackageNotFoundError:
            raise PackageNotFound(dist.egg_name)

        # the stdlib library (!!!) changed its behavior in Python 3.10 :(
        # https://docs.python.org/3.10/library/importlib.metadata.html#entry-points
        if hasattr(importlib.metadata, 'EntryPoints'):
            x = [
                (
                    ep.name,
                    entrypoint(
                        module_name=ep.module,
                        attrs=ep.attr.split('.'),
                    ),
                )
                for ep in dist.entry_points.select(group=group)
            ]
            return x
        else:
            x = [
                (
                    ep.name,
                    entrypoint(
                        module_name=ep.value.split(':')[0],
                        attrs=ep.value.split(':')[1].split('.'),
                    ),
                )
                for ep in dist.entry_points
                if ep.group == group
            ]
            return x
    elif metadata_lib == METADATA_LIB_BACKPORT:
        import importlib_metadata

        try:
            dist = importlib_metadata.Distribution.at(dist.egg_info)
        except importlib_metadata.PackageNotFoundError:
            raise PackageNotFound(dist.egg_name)

        # as above
        if hasattr(importlib_metadata, 'EntryPoints'):
            x = [
                (
                    ep.name,
                    entrypoint(
                        module_name=ep.module,
                        attrs=ep.attr.split('.'),
                    ),
                )
                for ep in dist.entry_points.select(group=group)
            ]
            return x
        else:
            x = [
                (
                    ep.name,
                    entrypoint(
                        module_name=ep.value.split(':')[0],
                        attrs=ep.value.split(':')[1].split('.'),
                    ),
                )
                for ep in dist.entry_points
                if ep.group == group
            ]
            return x
    else:  # METADATA_LIB_LEGACY
        import pkg_resources

        try:
            dist = pkg_resources.Distribution(
                dist.egg_base,
                pkg_resources.PathMetadata(dist.egg_base, dist.egg_info),
                dist.egg_name,
                dist.egg_version,
            )
        except pkg_resources.DistributionNotFound:
            raise PackageNotFound(dist.egg_name)

        return [
            (name, entrypoint(module_name=ep.module_name, attrs=ep.attrs))
            for name, ep in dist.get_entry_map(group).items()
        ]



# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright (C) 2013 Association of Universities for Research in Astronomy
#                    (AURA)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided
#        with the distribution.
#
#     3. The name of AURA and its representatives may not be used to
#        endorse or promote products derived from this software without
#        specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

# The code in this module is mostly copy/pasted out of the distutils2 source
# code, as recommended by Tarek Ziade.

from __future__ import absolute_import
from __future__ import print_function

# These first two imports are not used, but are needed to get around an
# irritating Python bug that can crop up when using ./setup.py test.
# See: http://www.eby-sarna.com/pipermail/peak/2010-May/003355.html
try:
    import multiprocessing  # noqa
except ImportError:
    pass
import logging  # noqa

import io
import os
import re
import shlex
import sys
import traceback
import warnings

from distutils import errors
from distutils import log
import setuptools
from setuptools import dist as st_dist
from setuptools import extension

from pbr._compat.five import ConfigParser
from pbr._compat.five import integer_types
from pbr._compat.five import string_type
from pbr._compat import packaging as packaging_compat
from pbr import extra_files
from pbr import hooks

"""Implementation of setup.cfg support."""

# A simplified RE for this; just checks that the line ends with version
# predicates in ()
_VERSION_SPEC_RE = re.compile(r'\s*(.*?)\s*\((.*)\)\s*$')

# Mappings from setup.cfg options, in (section, option) form, to setup()
# keyword arguments
CFG_TO_PY_SETUP_ARGS = (
    (('metadata', 'name'), 'name'),
    (('metadata', 'version'), 'version'),
    (('metadata', 'author'), 'author'),
    (('metadata', 'author_email'), 'author_email'),
    (('metadata', 'maintainer'), 'maintainer'),
    (('metadata', 'maintainer_email'), 'maintainer_email'),
    (('metadata', 'home_page'), 'url'),
    (('metadata', 'project_urls'), 'project_urls'),
    (('metadata', 'summary'), 'description'),
    (('metadata', 'keywords'), 'keywords'),
    (('metadata', 'description'), 'long_description'),
    (
        ('metadata', 'description_content_type'),
        'long_description_content_type',
    ),
    (('metadata', 'download_url'), 'download_url'),
    (('metadata', 'classifier'), 'classifiers'),
    (('metadata', 'platform'), 'platforms'),  # **
    (('metadata', 'license'), 'license'),
    # Use setuptools install_requires, not
    # broken distutils requires
    (('metadata', 'requires_dist'), 'install_requires'),
    (('metadata', 'setup_requires_dist'), 'setup_requires'),
    (('metadata', 'python_requires'), 'python_requires'),
    (('metadata', 'requires_python'), 'python_requires'),
    (('metadata', 'provides_dist'), 'provides'),  # **
    (('metadata', 'provides_extras'), 'provides_extras'),
    (('metadata', 'obsoletes_dist'), 'obsoletes'),  # **
    (('files', 'packages_root'), 'package_dir'),
    (('files', 'packages'), 'packages'),
    (('files', 'package_data'), 'package_data'),
    (('files', 'namespace_packages'), 'namespace_packages'),
    (('files', 'data_files'), 'data_files'),
    (('files', 'scripts'), 'scripts'),
    (('files', 'modules'), 'py_modules'),  # **
    (('global', 'commands'), 'cmdclass'),
    # Not supported in distutils2, but provided for
    # backwards compatibility with setuptools
    (('backwards_compat', 'zip_safe'), 'zip_safe'),
    (('backwards_compat', 'tests_require'), 'tests_require'),
    (('backwards_compat', 'dependency_links'), 'dependency_links'),
    (('backwards_compat', 'include_package_data'), 'include_package_data'),
)

DEPRECATED_CFG = {
    ('metadata', 'home_page'): (
        "Use '[metadata] url' (setup.cfg) or '[project.urls]' "
        "(pyproject.toml) instead"
    ),
    ('metadata', 'summary'): (
        "Use '[metadata] description' (setup.cfg) or '[project] description' "
        "(pyproject.toml) instead"
    ),
    ('metadata', 'description_file'): (
        "Use '[metadata] long_description' (setup.cfg) or '[project] readme' "
        "(pyproject.toml) instead"
    ),
    ('metadata', 'classifier'): (
        "Use '[metadata] classifiers' (setup.cfg) or '[project] classifiers' "
        "(pyproject.toml) instead"
    ),
    ('metadata', 'platform'): (
        "Use '[metadata] platforms' (setup.cfg) or "
        "'[tool.setuptools] platforms' (pyproject.toml) instead"
    ),
    ('metadata', 'requires_dist'): (
        "Use '[options] install_requires' (setup.cfg) or "
        "'[project] dependencies' (pyproject.toml) instead"
    ),
    ('metadata', 'setup_requires_dist'): (
        "Use '[options] setup_requires' (setup.cfg) or "
        "'[build-system] requires' (pyproject.toml) instead"
    ),
    ('metadata', 'python_requires'): (
        "Use '[options] python_requires' (setup.cfg) or "
        "'[project] requires-python' (pyproject.toml) instead"
    ),
    ('metadata', 'requires_python'): (
        "Use '[options] python_requires' (setup.cfg) or "
        "'[project] requires-python' (pyproject.toml) instead"
    ),
    ('metadata', 'provides_dist'): "This option is ignored by pip",
    ('metadata', 'provides_extras'): "This option is ignored by pip",
    ('metadata', 'obsoletes_dist'): "This option is ignored by pip",
    ('files', 'packages_root'): (
        "Use '[options] package_dir' (setup.cfg) or '[tools.setuptools] "
        "package_dir' (pyproject.toml) instead"
    ),
    ('files', 'packages'): (
        "Use '[options] packages' (setup.cfg) or '[tools.setuptools] "
        "packages' (pyproject.toml) instead"
    ),
    ('files', 'package_data'): (
        "Use '[options.package_data]' (setup.cfg) or "
        "'[tool.setuptools.package-data]' (pyproject.toml) instead"
    ),
    ('files', 'namespace_packages'): (
        "Use '[options] namespace_packages' (setup.cfg) or migrate to PEP "
        "420-style namespace packages instead"
    ),
    ('files', 'data_files'): (
        "For package data files, use '[options] package_data' (setup.cfg) "
        "or '[tools.setuptools] package_data' (pyproject.toml) instead. "
        "Support for non-package data files is deprecated in setuptools "
        "and their use is discouraged. If necessary, use "
        "'[options] data_files' (setup.cfg) or '[tools.setuptools] data-files'"
        "(pyproject.toml) instead."
    ),
    ('files', 'scripts'): (
        "Migrate to using the console_scripts entrypoint and use "
        "'[options.entry_points]' (setup.cfg) or '[project.scripts]' "
        "(pyproject.toml) instead"
    ),
    ('files', 'modules'): (
        "Use '[options] py_modules' (setup.cfg) or '[tools.setuptools] "
        "py-modules' (pyproject.toml) instead"
    ),
    ('backwards_compat', 'zip_safe'): (
        "This option is obsolete as it was only relevant in the context of "
        "eggs"
    ),
    ('backwards_compat', 'dependency_links'): (
        "This option is ignored by pip starting from pip 19.0"
    ),
    ('backwards_compat', 'tests_require'): (
        "This option is ignored by pip starting from pip 19.0"
    ),
    ('backwards_compat', 'include_package_data'): (
        "Use '[options] include_package_data' (setup.cfg) or "
        "'[tools.setuptools] include-package-data' (pyproject.toml) instead"
    ),
}

# setup() arguments that can have multiple values in setup.cfg
MULTI_FIELDS = (
    "classifiers",
    "platforms",
    "install_requires",
    "provides",
    "obsoletes",
    "namespace_packages",
    "packages",
    "package_data",
    "data_files",
    "scripts",
    "py_modules",
    "dependency_links",
    "setup_requires",
    "tests_require",
    "keywords",
    "cmdclass",
    "provides_extras",
)

# a mapping of removed keywords to the version of setuptools that they were deprecated in
REMOVED_KEYWORDS = {
    # https://setuptools.pypa.io/en/stable/history.html#v72-0-0
    'tests_requires': '72.0.0',
}

# setup() arguments that can have mapping values in setup.cfg
MAP_FIELDS = ("project_urls",)

# setup() arguments that contain boolean values
BOOL_FIELDS = ("zip_safe", "include_package_data")


def shlex_split(path):
    if os.name == 'nt':
        # shlex cannot handle paths that contain backslashes, treating those
        # as escape characters.
        path = path.replace("\\", "/")
        return [x.replace("/", "\\") for x in shlex.split(path)]

    return shlex.split(path)


def resolve_name(name):
    """Resolve a name like ``module.object`` to an object and return it.

    Raise ImportError if the module or name is not found.
    """
    parts = name.split('.')
    cursor = len(parts) - 1
    module_name = parts[:cursor]
    attr_name = parts[-1]

    while cursor > 0:
        try:
            ret = __import__('.'.join(module_name), fromlist=[attr_name])
            break
        except ImportError:
            if cursor == 0:
                raise
            cursor -= 1
            module_name = parts[:cursor]
            attr_name = parts[cursor]
            ret = ''

    for part in parts[cursor:]:
        try:
            ret = getattr(ret, part)
        except AttributeError:
            raise ImportError(name)

    return ret


def setup_cfg_to_args(path='setup.cfg', script_args=None):
    """Parse setup.cfg file.

    Parse a setup.cfg file and tranform pbr-specific options to the underlying
    setuptools opts.

    :param path: The setup.cfg path.
    :param script_args: List of commands setup.py was called with.
    :returns: A dictionary of kwargs to set on the underlying Distribution
        object.
    :raises DistutilsFileError: When the setup.cfg file is not found.
    """
    if script_args is None:
        script_args = ()

    # The method source code really starts here.
    parser = ConfigParser()

    if not os.path.exists(path):
        raise errors.DistutilsFileError(
            "file '%s' does not exist" % os.path.abspath(path)
        )

    try:
        parser.read(path, encoding='utf-8')
    except TypeError:
        # Python 2 doesn't accept the encoding kwarg
        parser.read(path)

    config = {}
    for section in parser.sections():
        config[section] = {}
        for k, value in parser.items(section):
            config[section][k.replace('-', '_')] = value

    # Run setup_hooks, if configured
    setup_hooks = has_get_option(config, 'global', 'setup_hooks')
    package_dir = has_get_option(config, 'files', 'packages_root')

    # Add the source package directory to sys.path in case it contains
    # additional hooks, and to make sure it's on the path before any existing
    # installations of the package
    if package_dir:
        package_dir = os.path.abspath(package_dir)
        sys.path.insert(0, package_dir)

    try:
        if setup_hooks:
            setup_hooks = [
                hook
                for hook in split_multiline(setup_hooks)
                if hook != 'pbr.hooks.setup_hook'
            ]
            for hook in setup_hooks:
                hook_fn = resolve_name(hook)
                try:
                    hook_fn(config)
                except SystemExit:
                    log.error('setup hook %s terminated the installation')
                except Exception:
                    e = sys.exc_info()[1]
                    log.error(
                        'setup hook %s raised exception: %s\n' % (hook, e)
                    )
                    log.error(traceback.format_exc())
                    sys.exit(1)

        # Run the pbr hook
        hooks.setup_hook(config)

        kwargs = setup_cfg_to_setup_kwargs(config, script_args)

        # Set default config overrides
        kwargs['include_package_data'] = True
        kwargs['zip_safe'] = False

        if has_get_option(config, 'global', 'compilers'):
            warnings.warn(
                'Support for custom compilers was removed in pbr 7.0 and the '
                '\'[global] compilers\' option is now ignored.',
                DeprecationWarning,
            )

        ext_modules = get_extension_modules(config)
        if ext_modules:
            kwargs['ext_modules'] = ext_modules

        entry_points = get_entry_points(config)
        if entry_points:
            kwargs['entry_points'] = entry_points

        # Handle the [files]/extra_files option
        files_extra_files = has_get_option(config, 'files', 'extra_files')
        if files_extra_files:
            extra_files.set_extra_files(split_multiline(files_extra_files))

    finally:
        # Perform cleanup if any paths were added to sys.path
        if package_dir:
            sys.path.pop(0)

    return kwargs


def _read_description_file(config):
    """Handle the legacy 'description_file' option."""
    long_description = has_get_option(config, 'metadata', 'long_description')
    if long_description:
        # if we have a long_description then do nothing: setuptools will take
        # care of this for us
        return None

    description_files = has_get_option(config, 'metadata', 'description_file')
    if not description_files:
        return None

    description_files = split_multiline(description_files)

    data = ''
    for filename in description_files:
        description_file = io.open(filename, encoding='utf-8')
        try:
            data += description_file.read().strip() + '\n\n'
        finally:
            description_file.close()

    return data


def setup_cfg_to_setup_kwargs(config, script_args=None):
    """Convert config options to kwargs.

    Processes the setup.cfg options and converts them to arguments accepted
    by setuptools' setup() function.
    """
    if script_args is None:
        script_args = ()

    kwargs = {}

    # Temporarily holds install_requires and extra_requires while we
    # parse env_markers.
    all_requirements = {}

    # We want people to use description and long_description over summary and
    # description but there is obvious overlap. If we see the both of the
    # former being used, don't normalize
    skip_description_normalization = False
    if has_get_option(config, 'metadata', 'description') and (
        has_get_option(config, 'metadata', 'long_description')
        or has_get_option(config, 'metadata', 'description_file')
    ):
        kwargs['description'] = has_get_option(
            config, 'metadata', 'description'
        )
        long_description = _read_description_file(config)
        if long_description:
            kwargs['long_description'] = long_description

        skip_description_normalization = True

    for alias, arg in CFG_TO_PY_SETUP_ARGS:
        section, option = alias

        if skip_description_normalization and alias in (
            ('metadata', 'summary'),
            ('metadata', 'description'),
        ):
            continue

        in_cfg_value = has_get_option(config, section, option)

        if alias == ('metadata', 'description') and not in_cfg_value:
            in_cfg_value = _read_description_file(config)

        if not in_cfg_value:
            continue

        if alias in DEPRECATED_CFG:
            warnings.warn(
                "The '[%s] %s' option is deprecated: %s"
                % (alias[0], alias[1], DEPRECATED_CFG[alias]),
                DeprecationWarning,
            )

        if arg in MULTI_FIELDS:
            in_cfg_value = split_multiline(in_cfg_value)
        elif arg in MAP_FIELDS:
            in_cfg_map = {}
            for i in split_multiline(in_cfg_value):
                k, v = i.split('=', 1)
                in_cfg_map[k.strip()] = v.strip()
            in_cfg_value = in_cfg_map
        elif arg in BOOL_FIELDS:
            # Provide some flexibility here...
            if in_cfg_value.lower() in ('true', 't', '1', 'yes', 'y'):
                in_cfg_value = True
            else:
                in_cfg_value = False

        if in_cfg_value:
            if arg in REMOVED_KEYWORDS and (
                packaging_compat.parse_version(setuptools.__version__)
                >= packaging_compat.parse_version(REMOVED_KEYWORDS[arg])
            ):
                # deprecation warnings, if any, will already have been logged,
                # so simply skip this
                continue

            if arg in ('install_requires', 'tests_require'):
                # Replaces PEP345-style version specs with the sort expected by
                # setuptools
                in_cfg_value = [
                    _VERSION_SPEC_RE.sub(r'\1\2', pred)
                    for pred in in_cfg_value
                ]

            if arg == 'install_requires':
                # Split install_requires into package,env_marker tuples
                # These will be re-assembled later
                install_requires = []
                requirement_pattern = (
                    r'(?P<package>[^;]*);?(?P<env_marker>[^#]*?)(?:\s*#.*)?$'
                )
                for requirement in in_cfg_value:
                    m = re.match(requirement_pattern, requirement)
                    requirement_package = m.group('package').strip()
                    env_marker = m.group('env_marker').strip()
                    install_requires.append((requirement_package, env_marker))
                all_requirements[''] = install_requires
            elif arg == 'package_dir':
                in_cfg_value = {'': in_cfg_value}
            elif arg in ('package_data', 'data_files'):
                data_files = {}
                firstline = True
                prev = None
                for line in in_cfg_value:
                    if '=' in line:
                        key, value = line.split('=', 1)
                        key_unquoted = shlex_split(key.strip())[0]
                        key, value = (key_unquoted, value.strip())
                        if key in data_files:
                            # Multiple duplicates of the same package name;
                            # this is for backwards compatibility of the old
                            # format prior to d2to1 0.2.6.
                            prev = data_files[key]
                            prev.extend(shlex_split(value))
                        else:
                            prev = data_files[key.strip()] = shlex_split(value)
                    elif firstline:
                        raise errors.DistutilsOptionError(
                            'malformed package_data first line %r (misses '
                            '"=")' % line
                        )
                    else:
                        prev.extend(shlex_split(line.strip()))
                    firstline = False
                if arg == 'data_files':
                    # the data_files value is a pointlessly different structure
                    # from the package_data value
                    data_files = sorted(data_files.items())
                in_cfg_value = data_files
            elif arg == 'cmdclass':
                cmdclass = {}
                dist = st_dist.Distribution()
                for cls_name in in_cfg_value:
                    cls = resolve_name(cls_name)
                    cmd = cls(dist)
                    cmdclass[cmd.get_command_name()] = cls
                in_cfg_value = cmdclass

        kwargs[arg] = in_cfg_value

    # Transform requirements with embedded environment markers to
    # setuptools' supported marker-per-requirement format.
    #
    # install_requires are treated as a special case of extras, before
    # being put back in the expected place
    #
    # fred =
    #     foo:marker
    #     bar
    # -> {'fred': ['bar'], 'fred:marker':['foo']}

    if 'extras' in config:
        requirement_pattern = (
            r'(?P<package>[^:]*):?(?P<env_marker>[^#]*?)(?:\s*#.*)?$'
        )
        extras = config['extras']
        # Add contents of test-requirements, if any, into an extra named
        # 'test' if one does not already exist.
        if 'test' not in extras:
            from pbr import packaging

            extras['test'] = "\n".join(
                packaging.parse_requirements(packaging.TEST_REQUIREMENTS_FILES)
            ).replace(';', ':')

        for extra in extras:
            extra_requirements = []
            requirements = split_multiline(extras[extra])
            for requirement in requirements:
                m = re.match(requirement_pattern, requirement)
                extras_value = m.group('package').strip()
                env_marker = m.group('env_marker')
                extra_requirements.append((extras_value, env_marker))
            all_requirements[extra] = extra_requirements

    # Transform the full list of requirements into:
    # - install_requires, for those that have no extra and no
    #   env_marker
    # - named extras, for those with an extra name (which may include
    #   an env_marker)
    # - and as a special case, install_requires with an env_marker are
    #   treated as named extras where the name is the empty string

    extras_require = {}
    for req_group in all_requirements:
        for requirement, env_marker in all_requirements[req_group]:
            if env_marker:
                extras_key = '%s:(%s)' % (req_group, env_marker)
                # We do not want to poison wheel creation with locally
                # evaluated markers.  sdists always re-create the egg_info
                # and as such do not need guarded, and pip will never call
                # multiple setup.py commands at once.
                if 'bdist_wheel' not in script_args:
                    try:
                        if packaging_compat.evaluate_marker(
                            '(%s)' % env_marker
                        ):
                            extras_key = req_group
                    except SyntaxError:
                        log.error(
                            "Marker evaluation failed, see the following "
                            "error.  For more information see: "
                            "http://docs.openstack.org/"
                            "pbr/latest/user/using.html#environment-markers"
                        )
                        raise
            else:
                extras_key = req_group
            extras_require.setdefault(extras_key, []).append(requirement)

    kwargs['install_requires'] = extras_require.pop('', [])
    kwargs['extras_require'] = extras_require

    return kwargs


def get_extension_modules(config):
    """Handle extension modules"""

    EXTENSION_FIELDS = (
        "sources",
        "include_dirs",
        "define_macros",
        "undef_macros",
        "library_dirs",
        "libraries",
        "runtime_library_dirs",
        "extra_objects",
        "extra_compile_args",
        "extra_link_args",
        "export_symbols",
        "swig_opts",
        "depends",
    )

    ext_modules = []
    for section in config:
        if ':' in section:
            labels = section.split(':', 1)
        else:
            # Backwards compatibility for old syntax; don't use this though
            labels = section.split('=', 1)
        labels = [label.strip() for label in labels]
        if (len(labels) == 2) and (labels[0] == 'extension'):
            ext_args = {}
            for field in EXTENSION_FIELDS:
                value = has_get_option(config, section, field)
                # All extension module options besides name can have multiple
                # values
                if not value:
                    continue
                value = split_multiline(value)
                if field == 'define_macros':
                    macros = []
                    for macro in value:
                        macro = macro.split('=', 1)
                        if len(macro) == 1:
                            macro = (macro[0].strip(), None)
                        else:
                            macro = (macro[0].strip(), macro[1].strip())
                        macros.append(macro)
                    value = macros
                ext_args[field] = value
            if ext_args:
                if 'name' not in ext_args:
                    ext_args['name'] = labels[1]
                ext_modules.append(
                    extension.Extension(ext_args.pop('name'), **ext_args)
                )
    return ext_modules


def get_entry_points(config):
    """Process the [entry_points] section of setup.cfg."""

    if 'entry_points' not in config:
        return {}

    warnings.warn(
        "The 'entry_points' section has been deprecated in favour of the "
        "'[options.entry_points]' section (if using 'setup.cfg') or the "
        "'[project.scripts]' and/or '[project.entry-points.{name}]' sections "
        "(if using 'pyproject.toml')",
        DeprecationWarning,
    )

    return {
        option: split_multiline(value)
        for option, value in config['entry_points'].items()
    }


def has_get_option(config, section, option):
    if section in config and option in config[section]:
        return config[section][option]
    else:
        return False


def split_multiline(value):
    """Special behaviour when we have a multi line options"""
    value = [
        element
        for element in (line.strip() for line in value.split('\n'))
        if element and not element.startswith('#')
    ]
    return value


def split_csv(value):
    """Special behaviour when we have a comma separated options"""
    value = [
        element
        for element in (chunk.strip() for chunk in value.split(','))
        if element
    ]
    return value


def pbr(dist, attr, value):
    """Implements the pbr setup() keyword.

    When used, this should be the only keyword in your setup() aside from
    `setup_requires`.

    If given as a string, the value of pbr is assumed to be the relative path
    to the setup.cfg file to use.  Otherwise, if it evaluates to true, it
    simply assumes that pbr should be used, and the default 'setup.cfg' is
    used.

    This works by reading the setup.cfg file, parsing out the supported
    metadata and command options, and using them to rebuild the
    `DistributionMetadata` object and set the newly added command options.

    The reason for doing things this way is that a custom `Distribution` class
    will not play nicely with setup_requires; however, this implementation may
    not work well with distributions that do use a `Distribution` subclass.
    """

    # Distribution.finalize_options() is what calls this method. That means
    # there is potential for recursion here. Recursion seems to be an issue
    # particularly when using PEP517 build-system configs without
    # setup_requires in setup.py. We can avoid the recursion by setting
    # this canary so we don't repeat ourselves.
    if hasattr(dist, '_pbr_initialized'):
        return
    dist._pbr_initialized = True

    if not value:
        return

    if isinstance(value, string_type):
        path = os.path.abspath(value)
    else:
        path = os.path.abspath('setup.cfg')

    if not os.path.exists(path):
        raise errors.DistutilsFileError(
            'The setup.cfg file %s does not exist.' % path
        )

    # Converts the setup.cfg file to setup() arguments
    try:
        attrs = setup_cfg_to_args(path, dist.script_args)
    except Exception:
        e = sys.exc_info()[1]
        # NB: This will output to the console if no explicit logging has
        # been setup - but thats fine, this is a fatal distutils error, so
        # being pretty isn't the #1 goal.. being diagnosable is.
        logging.exception('Error parsing')
        raise errors.DistutilsSetupError(
            'Error parsing %s: %s: %s' % (path, e.__class__.__name__, e)
        )

    # There are some metadata fields that are only supported by
    # setuptools and not distutils, and hence are not in
    # dist.metadata.  We are OK to write these in.  For gory details
    # see
    #  https://github.com/pypa/setuptools/pull/1343
    _DISTUTILS_UNSUPPORTED_METADATA = (
        'long_description_content_type',
        'project_urls',
        'provides_extras',
    )

    # Repeat some of the Distribution initialization code with the newly
    # provided attrs
    if attrs:
        # Skips 'options' and 'licence' support which are rarely used; may
        # add back in later if demanded
        for key, val in attrs.items():
            if hasattr(dist.metadata, 'set_' + key):
                getattr(dist.metadata, 'set_' + key)(val)
            elif hasattr(dist.metadata, key):
                setattr(dist.metadata, key, val)
            elif hasattr(dist, key):
                setattr(dist, key, val)
            elif key in _DISTUTILS_UNSUPPORTED_METADATA:
                setattr(dist.metadata, key, val)
            else:
                msg = 'Unknown distribution option: %s' % repr(key)
                warnings.warn(msg)

    # Re-finalize the underlying Distribution
    try:
        super(dist.__class__, dist).finalize_options()
    except TypeError:
        # If dist is not declared as a new-style class (with object as
        # a subclass) then super() will not work on it. This is the case
        # for Python 2. In that case, fall back to doing this the ugly way
        dist.__class__.__bases__[-1].finalize_options(dist)

    # This bit comes out of distribute/setuptools
    if isinstance(dist.metadata.version, integer_types + (float,)):
        # Some people apparently take "version number" too literally :)
        dist.metadata.version = str(dist.metadata.version)



# Copyright (c) 2011 OpenStack Foundation
# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import
from __future__ import print_function

import os
import sys

import fixtures

from pbr._compat.five import BytesIO
from pbr import git
from pbr import options
from pbr.tests import base

if sys.version_info >= (3, 3):
    from unittest import mock
else:
    import mock  # noqa


class TestGitIsInstalled(base.BaseTestCase):

    def testGitIsInstalled(self):
        with mock.patch.object(git, '_run_shell_command') as _command:
            _command.return_value = 'git version 1.8.4.1'
            self.assertEqual(True, git._git_is_installed())

    def testGitIsNotInstalled(self):
        with mock.patch.object(git, '_run_shell_command') as _command:
            _command.side_effect = OSError
            self.assertEqual(False, git._git_is_installed())


class SkipFileWrites(base.BaseTestCase):

    scenarios = [
        (
            'changelog_option_true',
            {
                'option_key': 'skip_changelog',
                'option_value': 'True',
                'env_key': 'SKIP_WRITE_GIT_CHANGELOG',
                'env_value': None,
                'pkg_func': git.write_git_changelog,
                'filename': 'ChangeLog',
            },
        ),
        (
            'changelog_option_false',
            {
                'option_key': 'skip_changelog',
                'option_value': 'False',
                'env_key': 'SKIP_WRITE_GIT_CHANGELOG',
                'env_value': None,
                'pkg_func': git.write_git_changelog,
                'filename': 'ChangeLog',
            },
        ),
        (
            'changelog_env_true',
            {
                'option_key': 'skip_changelog',
                'option_value': 'False',
                'env_key': 'SKIP_WRITE_GIT_CHANGELOG',
                'env_value': 'True',
                'pkg_func': git.write_git_changelog,
                'filename': 'ChangeLog',
            },
        ),
        (
            'changelog_both_true',
            {
                'option_key': 'skip_changelog',
                'option_value': 'True',
                'env_key': 'SKIP_WRITE_GIT_CHANGELOG',
                'env_value': 'True',
                'pkg_func': git.write_git_changelog,
                'filename': 'ChangeLog',
            },
        ),
        (
            'authors_option_true',
            {
                'option_key': 'skip_authors',
                'option_value': 'True',
                'env_key': 'SKIP_GENERATE_AUTHORS',
                'env_value': None,
                'pkg_func': git.generate_authors,
                'filename': 'AUTHORS',
            },
        ),
        (
            'authors_option_false',
            {
                'option_key': 'skip_authors',
                'option_value': 'False',
                'env_key': 'SKIP_GENERATE_AUTHORS',
                'env_value': None,
                'pkg_func': git.generate_authors,
                'filename': 'AUTHORS',
            },
        ),
        (
            'authors_env_true',
            {
                'option_key': 'skip_authors',
                'option_value': 'False',
                'env_key': 'SKIP_GENERATE_AUTHORS',
                'env_value': 'True',
                'pkg_func': git.generate_authors,
                'filename': 'AUTHORS',
            },
        ),
        (
            'authors_both_true',
            {
                'option_key': 'skip_authors',
                'option_value': 'True',
                'env_key': 'SKIP_GENERATE_AUTHORS',
                'env_value': 'True',
                'pkg_func': git.generate_authors,
                'filename': 'AUTHORS',
            },
        ),
    ]

    def setUp(self):
        super(SkipFileWrites, self).setUp()
        self.temp_path = self.useFixture(fixtures.TempDir()).path
        self.root_dir = os.path.abspath(os.path.curdir)
        self.git_dir = os.path.join(self.root_dir, ".git")
        if not os.path.exists(self.git_dir):
            self.skipTest(
                "%s is missing; skipping git-related checks" % self.git_dir
            )
            return
        self.filename = os.path.join(self.temp_path, self.filename)
        self.option_dict = {}
        if self.option_key is not None:
            self.option_dict[self.option_key] = (
                'setup.cfg',
                self.option_value,
            )
        self.useFixture(
            fixtures.EnvironmentVariable(self.env_key, self.env_value)
        )

    def test_skip(self):
        self.pkg_func(
            git_dir=self.git_dir,
            dest_dir=self.temp_path,
            option_dict=self.option_dict,
        )
        self.assertEqual(
            not os.path.exists(self.filename),
            (
                self.option_value.lower() in options.TRUE_VALUES
                or self.env_value is not None
            ),
        )


_changelog_content = """7780758\x00Break parser\x00 (tag: refs/tags/1_foo.1)
04316fe\x00Make python\x00 (refs/heads/review/monty_taylor/27519)
378261a\x00Add an integration test script.\x00
3c373ac\x00Merge "Lib\x00 (HEAD, tag: refs/tags/2013.2.rc2, tag: refs/tags/2013.2, refs/heads/mile-proposed)
182feb3\x00Fix pip invocation for old versions of pip.\x00 (tag: refs/tags/0.5.17)
fa4f46e\x00Remove explicit depend on distribute.\x00 (tag: refs/tags/0.5.16)
d1c53dd\x00Use pip instead of easy_install for installation.\x00
a793ea1\x00Merge "Skip git-checkout related tests when .git is missing"\x00
6c27ce7\x00Skip git-checkout related tests when .git is missing\x00
451e513\x00Bug fix: create_stack() fails when waiting\x00
4c8cfe4\x00Improve test coverage: network delete API\x00 (tag: refs/tags/(evil))
d7e6167\x00Bug fix: Fix pass thru filtering in list_networks\x00 (tag: refs/tags/ev()il)
c47ec15\x00Consider 'in-use' a non-pending volume for caching\x00 (tag: refs/tags/ev)il)
8696fbd\x00Improve test coverage: private extension API\x00 (tag: refs/tags/ev(il)
f0440f8\x00Improve test coverage: hypervisor list\x00 (tag: refs/tags/e(vi)l)
04984a5\x00Refactor hooks file.\x00 (HEAD, tag: 0.6.7,b, tag: refs/tags/(12), refs/heads/master)
a65e8ee\x00Remove jinja pin.\x00 (tag: refs/tags/0.5.14, tag: refs/tags/0.5.13)
"""  # noqa


def _make_old_git_changelog_format(line):
    """Convert post-1.8.1 git log format to pre-1.8.1 git log format"""

    if not line.strip():
        return line
    sha, msg, refname = line.split('\x00')
    refname = refname.replace('tag: ', '')
    return '\x00'.join((sha, msg, refname))


_old_git_changelog_content = '\n'.join(
    _make_old_git_changelog_format(line)
    for line in _changelog_content.split('\n')
)


class GitLogsTest(base.BaseTestCase):

    scenarios = [
        ('pre1.8.3', {'changelog': _old_git_changelog_content}),
        ('post1.8.3', {'changelog': _changelog_content}),
    ]

    def setUp(self):
        super(GitLogsTest, self).setUp()
        self.temp_path = self.useFixture(fixtures.TempDir()).path
        self.root_dir = os.path.abspath(os.path.curdir)
        self.git_dir = os.path.join(self.root_dir, ".git")
        self.useFixture(fixtures.EnvironmentVariable('SKIP_GENERATE_AUTHORS'))
        self.useFixture(
            fixtures.EnvironmentVariable('SKIP_WRITE_GIT_CHANGELOG')
        )

    def test_write_git_changelog(self):
        self.useFixture(
            fixtures.FakePopen(
                lambda _: {"stdout": BytesIO(self.changelog.encode('utf-8'))}
            )
        )

        git.write_git_changelog(git_dir=self.git_dir, dest_dir=self.temp_path)

        with open(os.path.join(self.temp_path, "ChangeLog"), "r") as ch_fh:
            changelog_contents = ch_fh.read()
            self.assertIn("2013.2", changelog_contents)
            self.assertIn("0.5.17", changelog_contents)
            self.assertIn("------", changelog_contents)
            self.assertIn("Refactor hooks file", changelog_contents)
            self.assertIn(
                r"Bug fix: create\_stack() fails when waiting",
                changelog_contents,
            )
            self.assertNotIn("Refactor hooks file.", changelog_contents)
            self.assertNotIn("182feb3", changelog_contents)
            self.assertNotIn("review/monty_taylor/27519", changelog_contents)
            self.assertNotIn("0.5.13", changelog_contents)
            self.assertNotIn("0.6.7", changelog_contents)
            self.assertNotIn("12", changelog_contents)
            self.assertNotIn("(evil)", changelog_contents)
            self.assertNotIn("ev()il", changelog_contents)
            self.assertNotIn("ev(il", changelog_contents)
            self.assertNotIn("ev)il", changelog_contents)
            self.assertNotIn("e(vi)l", changelog_contents)
            self.assertNotIn('Merge "', changelog_contents)
            self.assertNotIn(r'1\_foo.1', changelog_contents)

    def test_generate_authors(self):
        author_old = u"Foo Foo <email@foo.com>"
        author_new = u"Bar Bar <email@bar.com>"
        co_author = u"Foo Bar <foo@bar.com>"
        co_author_by = u"Co-authored-by: " + co_author

        git_log_cmd = (
            "git --git-dir=%s log --format=%%aN <%%aE>" % self.git_dir
        )
        git_co_log_cmd = "git --git-dir=%s log" % self.git_dir
        git_top_level = "git rev-parse --show-toplevel"
        cmd_map = {
            git_log_cmd: author_new,
            git_co_log_cmd: co_author_by,
            git_top_level: self.root_dir,
        }

        exist_files = [
            self.git_dir,
            os.path.join(self.temp_path, "AUTHORS.in"),
        ]
        self.useFixture(
            fixtures.MonkeyPatch(
                "os.path.exists",
                lambda path: os.path.abspath(path) in exist_files,
            )
        )

        def _fake_run_shell_command(cmd, **kwargs):
            return cmd_map[" ".join(cmd)]

        self.useFixture(
            fixtures.MonkeyPatch(
                "pbr.git._run_shell_command", _fake_run_shell_command
            )
        )

        with open(os.path.join(self.temp_path, "AUTHORS.in"), "w") as auth_fh:
            auth_fh.write("%s\n" % author_old)

        git.generate_authors(git_dir=self.git_dir, dest_dir=self.temp_path)

        with open(os.path.join(self.temp_path, "AUTHORS"), "r") as auth_fh:
            authors = auth_fh.read()
            self.assertIn(author_old, authors)
            self.assertIn(author_new, authors)
            self.assertIn(co_author, authors)



# Copyright (c) 2013 New Dream Network, LLC (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright (C) 2013 Association of Universities for Research in Astronomy
#                    (AURA)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided
#        with the distribution.
#
#     3. The name of AURA and its representatives may not be used to
#        endorse or promote products derived from this software without
#        specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS

from __future__ import absolute_import
from __future__ import print_function

import email.errors
import os
import re
import sys
import tempfile

import fixtures
import testscenarios
from testtools import matchers

from pbr import packaging
from pbr.tests import base
from pbr.tests import fixtures as pbr_fixtures

if sys.version_info >= (3, 3):
    from unittest import mock
else:
    import mock  # noqa


class ParseRequirementsTest(base.BaseTestCase):

    def test_empty_requirements(self):
        actual = packaging.parse_requirements([])
        self.assertEqual([], actual)

    def test_default_requirements(self):
        """Ensure default files used if no files provided."""
        tempdir = tempfile.mkdtemp()
        requirements = os.path.join(tempdir, 'requirements.txt')
        with open(requirements, 'w') as f:
            f.write('pbr')
        # the defaults are relative to where pbr is called from so we need to
        # override them. This is OK, however, as we want to validate that
        # defaults are used - not what those defaults are
        with mock.patch.object(
            packaging, 'REQUIREMENTS_FILES', (requirements,)
        ):
            result = packaging.parse_requirements()
        self.assertEqual(['pbr'], result)

    def test_override_with_env(self):
        """Ensure environment variable used if no files provided."""
        _, tmp_file = tempfile.mkstemp(prefix='openstack', suffix='.setup')
        with open(tmp_file, 'w') as fh:
            fh.write("foo\nbar")
        self.useFixture(
            fixtures.EnvironmentVariable('PBR_REQUIREMENTS_FILES', tmp_file)
        )
        self.assertEqual(['foo', 'bar'], packaging.parse_requirements())

    def test_override_with_env_multiple_files(self):
        _, tmp_file = tempfile.mkstemp(prefix='openstack', suffix='.setup')
        with open(tmp_file, 'w') as fh:
            fh.write("foo\nbar")
        self.useFixture(
            fixtures.EnvironmentVariable(
                'PBR_REQUIREMENTS_FILES', "no-such-file," + tmp_file
            )
        )
        self.assertEqual(['foo', 'bar'], packaging.parse_requirements())

    def test_index_present(self):
        tempdir = tempfile.mkdtemp()
        requirements = os.path.join(tempdir, 'requirements.txt')
        with open(requirements, 'w') as f:
            f.write('-i https://myindex.local\n')
            f.write('  --index-url https://myindex.local\n')
            f.write(' --extra-index-url https://myindex.local\n')
            f.write('--find-links https://myindex.local\n')
            f.write('arequirement>=1.0\n')
        result = packaging.parse_requirements([requirements])
        self.assertEqual(['arequirement>=1.0'], result)

    def test_nested_requirements(self):
        tempdir = tempfile.mkdtemp()
        requirements = os.path.join(tempdir, 'requirements.txt')
        nested = os.path.join(tempdir, 'nested.txt')
        with open(requirements, 'w') as f:
            f.write('-r ' + nested)
        with open(nested, 'w') as f:
            f.write('pbr')
        result = packaging.parse_requirements([requirements])
        self.assertEqual(['pbr'], result)


class ParseRequirementsTestScenarios(base.BaseTestCase):

    versioned_scenarios = [
        ('non-versioned', {'versioned': False, 'expected': ['bar']}),
        ('versioned', {'versioned': True, 'expected': ['bar>=1.2.3']}),
    ]

    subdirectory_scenarios = [
        ('non-subdirectory', {'has_subdirectory': False}),
        ('has-subdirectory', {'has_subdirectory': True}),
    ]

    scenarios = [
        ('normal', {'url': "foo\nbar", 'expected': ['foo', 'bar']}),
        (
            'normal_with_comments',
            {
                'url': "# this is a comment\nfoo\n# and another one\nbar",
                'expected': ['foo', 'bar'],
            },
        ),
        ('removes_index_lines', {'url': '-f foobar', 'expected': []}),
    ]

    scenarios = scenarios + testscenarios.multiply_scenarios(
        [
            ('ssh_egg_url', {'url': 'git+ssh://foo.com/zipball#egg=bar'}),
            (
                'git_https_egg_url',
                {'url': 'git+https://foo.com/zipball#egg=bar'},
            ),
            ('http_egg_url', {'url': 'https://foo.com/zipball#egg=bar'}),
        ],
        versioned_scenarios,
        subdirectory_scenarios,
    )

    scenarios = scenarios + testscenarios.multiply_scenarios(
        [
            (
                'git_egg_url',
                {'url': 'git://foo.com/zipball#egg=bar', 'name': 'bar'},
            )
        ],
        [
            ('non-editable', {'editable': False}),
            ('editable', {'editable': True}),
        ],
        versioned_scenarios,
        subdirectory_scenarios,
    )

    def test_parse_requirements(self):
        tmp_file = tempfile.NamedTemporaryFile()
        req_string = self.url
        if hasattr(self, 'editable') and self.editable:
            req_string = "-e %s" % req_string
        if hasattr(self, 'versioned') and self.versioned:
            req_string = "%s-1.2.3" % req_string
        if hasattr(self, 'has_subdirectory') and self.has_subdirectory:
            req_string = "%s&subdirectory=baz" % req_string
        with open(tmp_file.name, 'w') as fh:
            fh.write(req_string)
        self.assertEqual(
            self.expected, packaging.parse_requirements([tmp_file.name])
        )


class ParseDependencyLinksTest(base.BaseTestCase):

    def setUp(self):
        super(ParseDependencyLinksTest, self).setUp()
        _, self.tmp_file = tempfile.mkstemp(
            prefix="openstack", suffix=".setup"
        )

    def test_parse_dependency_normal(self):
        with open(self.tmp_file, "w") as fh:
            fh.write("http://test.com\n")
        self.assertEqual(
            ["http://test.com"],
            packaging.parse_dependency_links([self.tmp_file]),
        )

    def test_parse_dependency_with_git_egg_url(self):
        with open(self.tmp_file, "w") as fh:
            fh.write("-e git://foo.com/zipball#egg=bar")
        self.assertEqual(
            ["git://foo.com/zipball#egg=bar"],
            packaging.parse_dependency_links([self.tmp_file]),
        )


class TestVersions(base.BaseTestCase):

    scenarios = [
        ('preversioned', {'preversioned': True}),
        ('postversioned', {'preversioned': False}),
    ]

    def setUp(self):
        super(TestVersions, self).setUp()
        self.repo = self.useFixture(pbr_fixtures.GitRepo(self.package_dir))
        self.useFixture(pbr_fixtures.GPGKey())
        self.useFixture(pbr_fixtures.Chdir(self.package_dir))

    def tearDown(self):
        super(TestVersions, self).tearDown()
        os.environ.pop('SKIP_WRITE_GIT_CHANGELOG', None)

    def test_email_parsing_errors_are_handled(self):
        mocked_open = mock.mock_open()
        with mock.patch('pbr.packaging.open', mocked_open):
            with mock.patch('email.message_from_file') as message_from_file:
                message_from_file.side_effect = [
                    email.errors.MessageError('Test'),
                    {'Name': 'pbr_testpackage'},
                ]
                version = packaging._get_version_from_pkg_metadata(
                    'pbr_testpackage'
                )

        self.assertTrue(message_from_file.called)
        self.assertIsNone(version)

    def test_capitalized_headers(self):
        self.repo.commit()
        self.repo.tag('1.2.3')
        self.repo.commit('Sem-Ver: api-break')
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('2.0.0.dev1'))

    def test_capitalized_headers_partial(self):
        self.repo.commit()
        self.repo.tag('1.2.3')
        self.repo.commit('Sem-ver: api-break')
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('2.0.0.dev1'))

    def test_multi_inline_symbols_no_space(self):
        self.repo.commit()
        self.repo.tag('1.2.3')
        self.repo.commit('Sem-ver: feature,api-break')
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('2.0.0.dev1'))

    def test_multi_inline_symbols_spaced(self):
        self.repo.commit()
        self.repo.tag('1.2.3')
        self.repo.commit('Sem-ver: feature, api-break')
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('2.0.0.dev1'))

    def test_multi_inline_symbols_reversed(self):
        self.repo.commit()
        self.repo.tag('1.2.3')
        self.repo.commit('Sem-ver: api-break,feature')
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('2.0.0.dev1'))

    def test_leading_space(self):
        self.repo.commit()
        self.repo.tag('1.2.3')
        self.repo.commit('   sem-ver: api-break')
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('2.0.0.dev1'))

    def test_leading_space_multiline(self):
        self.repo.commit()
        self.repo.tag('1.2.3')
        self.repo.commit(('   Some cool text\n   sem-ver: api-break'))
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('2.0.0.dev1'))

    def test_leading_characters_symbol_not_found(self):
        self.repo.commit()
        self.repo.tag('1.2.3')
        self.repo.commit('  ssem-ver: api-break')
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('1.2.4.dev1'))

    def test_tagged_version_has_tag_version(self):
        self.repo.commit()
        self.repo.tag('1.2.3')
        version = packaging._get_version_from_git('1.2.3')
        self.assertEqual('1.2.3', version)

    def test_tagged_version_with_semver_compliant_prerelease(self):
        self.repo.commit()
        self.repo.tag('1.2.3-rc2')
        version = packaging._get_version_from_git()
        self.assertEqual('1.2.3.0rc2', version)

    def test_non_canonical_tagged_version_bump(self):
        self.repo.commit()
        self.repo.tag('1.4')
        self.repo.commit('Sem-Ver: api-break')
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('2.0.0.dev1'))

    def test_untagged_version_has_dev_version_postversion(self):
        self.repo.commit()
        self.repo.tag('1.2.3')
        self.repo.commit()
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('1.2.4.dev1'))

    def test_untagged_pre_release_has_pre_dev_version_postversion(self):
        self.repo.commit()
        self.repo.tag('1.2.3.0a1')
        self.repo.commit()
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('1.2.3.0a2.dev1'))

    def test_untagged_version_minor_bump(self):
        self.repo.commit()
        self.repo.tag('1.2.3')
        self.repo.commit('sem-ver: deprecation')
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('1.3.0.dev1'))

    def test_untagged_version_major_bump(self):
        self.repo.commit()
        self.repo.tag('1.2.3')
        self.repo.commit('sem-ver: api-break')
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('2.0.0.dev1'))

    def test_untagged_version_has_dev_version_preversion(self):
        self.repo.commit()
        self.repo.tag('1.2.3')
        self.repo.commit()
        version = packaging._get_version_from_git('1.2.5')
        self.assertThat(version, matchers.StartsWith('1.2.5.dev1'))

    def test_untagged_version_after_pre_has_dev_version_preversion(self):
        self.repo.commit()
        self.repo.tag('1.2.3.0a1')
        self.repo.commit()
        version = packaging._get_version_from_git('1.2.5')
        self.assertThat(version, matchers.StartsWith('1.2.5.dev1'))

    def test_untagged_version_after_rc_has_dev_version_preversion(self):
        self.repo.commit()
        self.repo.tag('1.2.3.0a1')
        self.repo.commit()
        version = packaging._get_version_from_git('1.2.3')
        self.assertThat(version, matchers.StartsWith('1.2.3.0a2.dev1'))

    def test_untagged_version_after_semver_compliant_prerelease_tag(self):
        self.repo.commit()
        self.repo.tag('1.2.3-rc2')
        self.repo.commit()
        version = packaging._get_version_from_git()
        self.assertEqual('1.2.3.0rc3.dev1', version)

    def test_preversion_too_low_simple(self):
        # That is, the target version is either already released or not high
        # enough for the semver requirements given api breaks etc.
        self.repo.commit()
        self.repo.tag('1.2.3')
        self.repo.commit()
        # Note that we can't target 1.2.3 anymore - with 1.2.3 released we
        # need to be working on 1.2.4.
        err = self.assertRaises(
            ValueError, packaging._get_version_from_git, '1.2.3'
        )
        self.assertThat(err.args[0], matchers.StartsWith('git history'))

    def test_preversion_too_low_semver_headers(self):
        # That is, the target version is either already released or not high
        # enough for the semver requirements given api breaks etc.
        self.repo.commit()
        self.repo.tag('1.2.3')
        self.repo.commit('sem-ver: feature')
        # Note that we can't target 1.2.4, the feature header means we need
        # to be working on 1.3.0 or above.
        err = self.assertRaises(
            ValueError, packaging._get_version_from_git, '1.2.4'
        )
        self.assertThat(err.args[0], matchers.StartsWith('git history'))

    def test_get_kwargs_corner_cases(self):
        # No tags:

        def get_kwargs(tag):
            git_dir = self.repo._basedir + '/.git'
            return packaging._get_increment_kwargs(git_dir, tag)

        def _check_combinations(tag):
            self.repo.commit()
            self.assertEqual({}, get_kwargs(tag))
            self.repo.commit('sem-ver: bugfix')
            self.assertEqual({}, get_kwargs(tag))
            self.repo.commit('sem-ver: feature')
            self.assertEqual({'minor': True}, get_kwargs(tag))
            self.repo.uncommit()
            self.repo.commit('sem-ver: deprecation')
            self.assertEqual({'minor': True}, get_kwargs(tag))
            self.repo.uncommit()
            self.repo.commit('sem-ver: api-break')
            self.assertEqual({'major': True}, get_kwargs(tag))
            self.repo.commit('sem-ver: deprecation')
            self.assertEqual({'major': True, 'minor': True}, get_kwargs(tag))

        _check_combinations('')
        self.repo.tag('1.2.3')
        _check_combinations('1.2.3')

    def test_invalid_tag_ignored(self):
        # Fix for bug 1356784 - we treated any tag as a version, not just those
        # that are valid versions.
        self.repo.commit()
        self.repo.tag('1')
        self.repo.commit()
        # when the tree is tagged and its wrong:
        self.repo.tag('badver')
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('1.0.1.dev1'))
        # When the tree isn't tagged, we also fall through.
        self.repo.commit()
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('1.0.1.dev2'))
        # We don't fall through x.y versions
        self.repo.commit()
        self.repo.tag('1.2')
        self.repo.commit()
        self.repo.tag('badver2')
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('1.2.1.dev1'))
        # Or x.y.z versions
        self.repo.commit()
        self.repo.tag('1.2.3')
        self.repo.commit()
        self.repo.tag('badver3')
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('1.2.4.dev1'))
        # Or alpha/beta/pre versions
        self.repo.commit()
        self.repo.tag('1.2.4.0a1')
        self.repo.commit()
        self.repo.tag('badver4')
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('1.2.4.0a2.dev1'))
        # Non-release related tags are ignored.
        self.repo.commit()
        self.repo.tag('2')
        self.repo.commit()
        self.repo.tag('non-release-tag/2014.12.16-1')
        version = packaging._get_version_from_git()
        self.assertThat(version, matchers.StartsWith('2.0.1.dev1'))

    def test_valid_tag_honoured(self):
        # Fix for bug 1370608 - we converted any target into a 'dev version'
        # even if there was a distance of 0 - indicating that we were on the
        # tag itself.
        self.repo.commit()
        self.repo.tag('1.3.0.0a1')
        version = packaging._get_version_from_git()
        self.assertEqual('1.3.0.0a1', version)

    def test_skip_write_git_changelog(self):
        # Fix for bug 1467440
        self.repo.commit()
        self.repo.tag('1.2.3')
        os.environ['SKIP_WRITE_GIT_CHANGELOG'] = '1'
        version = packaging._get_version_from_git('1.2.3')
        self.assertEqual('1.2.3', version)


class TestRepositoryURLDependencies(base.BaseTestCase):

    def setUp(self):
        super(TestRepositoryURLDependencies, self).setUp()
        self.requirements = os.path.join(
            tempfile.mkdtemp(), 'requirements.txt'
        )
        with open(self.requirements, 'w') as f:
            f.write(
                '\n'.join(
                    [
                        '-e git+git://git.pro-ject.org/oslo.messaging#egg=oslo.messaging-1.0.0-rc',  # noqa
                        '-e git+git://git.pro-ject.org/django-thumborize#egg=django-thumborize',  # noqa
                        '-e git+git://git.pro-ject.org/django-thumborize#egg=django-thumborize-beta',  # noqa
                        '-e git+git://git.pro-ject.org/django-thumborize#egg=django-thumborize2-beta',  # noqa
                        '-e git+git://git.pro-ject.org/django-thumborize#egg=django-thumborize2-beta-4.0.1',  # noqa
                        '-e git+git://git.pro-ject.org/django-thumborize#egg=django-thumborize2-beta-1.0.0-alpha.beta.1',  # noqa
                        '-e git+git://git.pro-ject.org/django-thumborize#egg=django-thumborize2-beta-1.0.0-alpha-a.b-c-somethinglong+build.1-aef.1-its-okay',  # noqa
                        '-e git+git://git.pro-ject.org/django-thumborize#egg=django-thumborize2-beta-2.0.0-rc.1+build.123',  # noqa
                        '-e git+git://git.project.org/Proj#egg=Proj1',
                        'git+https://git.project.org/Proj#egg=Proj2-0.0.1',
                        '-e git+ssh://git.project.org/Proj#egg=Proj3',
                        'svn+svn://svn.project.org/svn/Proj#egg=Proj4-0.0.2',
                        '-e svn+http://svn.project.org/svn/Proj/trunk@2019#egg=Proj5',
                        'hg+http://hg.project.org/Proj@da39a3ee5e6b#egg=Proj-0.0.3',
                        '-e hg+http://hg.project.org/Proj@2019#egg=Proj',
                        'hg+http://hg.project.org/Proj@v1.0#egg=Proj-0.0.4',
                        '-e hg+http://hg.project.org/Proj@special_feature#egg=Proj',
                        'git://foo.com/zipball#egg=foo-bar-1.2.4',
                        'pypi-proj1',
                        'pypi-proj2',
                    ]
                )
            )

    def test_egg_fragment(self):
        expected = [
            'django-thumborize',
            'django-thumborize-beta',
            'django-thumborize2-beta',
            'django-thumborize2-beta>=4.0.1',
            'django-thumborize2-beta>=1.0.0-alpha.beta.1',
            'django-thumborize2-beta>=1.0.0-alpha-a.b-c-long+build.1-aef.1-its-okay',  # noqa
            'django-thumborize2-beta>=2.0.0-rc.1+build.123',
            'django-thumborize-beta>=0.0.4',
            'django-thumborize-beta>=1.2.3',
            'django-thumborize-beta>=10.20.30',
            'django-thumborize-beta>=1.1.2-prerelease+meta',
            'django-thumborize-beta>=1.1.2+meta',
            'django-thumborize-beta>=1.1.2+meta-valid',
            'django-thumborize-beta>=1.0.0-alpha',
            'django-thumborize-beta>=1.0.0-beta',
            'django-thumborize-beta>=1.0.0-alpha.beta',
            'django-thumborize-beta>=1.0.0-alpha.beta.1',
            'django-thumborize-beta>=1.0.0-alpha.1',
            'django-thumborize-beta>=1.0.0-alpha0.valid',
            'django-thumborize-beta>=1.0.0-alpha.0valid',
            'django-thumborize-beta>=1.0.0-alpha-a.b-c-somethinglong+build.1-aef.1-its-okay',  # noqa
            'django-thumborize-beta>=1.0.0-rc.1+build.1',
            'django-thumborize-beta>=2.0.0-rc.1+build.123',
            'django-thumborize-beta>=1.2.3-beta',
            'django-thumborize-beta>=10.2.3-DEV-SNAPSHOT',
            'django-thumborize-beta>=1.2.3-SNAPSHOT-123',
            'django-thumborize-beta>=1.0.0',
            'django-thumborize-beta>=2.0.0',
            'django-thumborize-beta>=1.1.7',
            'django-thumborize-beta>=2.0.0+build.1848',
            'django-thumborize-beta>=2.0.1-alpha.1227',
            'django-thumborize-beta>=1.0.0-alpha+beta',
            'django-thumborize-beta>=1.2.3----RC-SNAPSHOT.12.9.1--.12+788',
            'django-thumborize-beta>=1.2.3----R-S.12.9.1--.12+meta',
            'django-thumborize-beta>=1.2.3----RC-SNAPSHOT.12.9.1--.12',
            'django-thumborize-beta>=1.0.0+0.build.1-rc.10000aaa-kk-0.1',
            'django-thumborize-beta>=999999999999999999.99999999999999.9999999999999',  # noqa
            'Proj1',
            'Proj2>=0.0.1',
            'Proj3',
            'Proj4>=0.0.2',
            'Proj5',
            'Proj>=0.0.3',
            'Proj',
            'Proj>=0.0.4',
            'Proj',
            'foo-bar>=1.2.4',
        ]
        tests = [
            'egg=django-thumborize',
            'egg=django-thumborize-beta',
            'egg=django-thumborize2-beta',
            'egg=django-thumborize2-beta-4.0.1',
            'egg=django-thumborize2-beta-1.0.0-alpha.beta.1',
            'egg=django-thumborize2-beta-1.0.0-alpha-a.b-c-long+build.1-aef.1-its-okay',  # noqa
            'egg=django-thumborize2-beta-2.0.0-rc.1+build.123',
            'egg=django-thumborize-beta-0.0.4',
            'egg=django-thumborize-beta-1.2.3',
            'egg=django-thumborize-beta-10.20.30',
            'egg=django-thumborize-beta-1.1.2-prerelease+meta',
            'egg=django-thumborize-beta-1.1.2+meta',
            'egg=django-thumborize-beta-1.1.2+meta-valid',
            'egg=django-thumborize-beta-1.0.0-alpha',
            'egg=django-thumborize-beta-1.0.0-beta',
            'egg=django-thumborize-beta-1.0.0-alpha.beta',
            'egg=django-thumborize-beta-1.0.0-alpha.beta.1',
            'egg=django-thumborize-beta-1.0.0-alpha.1',
            'egg=django-thumborize-beta-1.0.0-alpha0.valid',
            'egg=django-thumborize-beta-1.0.0-alpha.0valid',
            'egg=django-thumborize-beta-1.0.0-alpha-a.b-c-somethinglong+build.1-aef.1-its-okay',  # noqa
            'egg=django-thumborize-beta-1.0.0-rc.1+build.1',
            'egg=django-thumborize-beta-2.0.0-rc.1+build.123',
            'egg=django-thumborize-beta-1.2.3-beta',
            'egg=django-thumborize-beta-10.2.3-DEV-SNAPSHOT',
            'egg=django-thumborize-beta-1.2.3-SNAPSHOT-123',
            'egg=django-thumborize-beta-1.0.0',
            'egg=django-thumborize-beta-2.0.0',
            'egg=django-thumborize-beta-1.1.7',
            'egg=django-thumborize-beta-2.0.0+build.1848',
            'egg=django-thumborize-beta-2.0.1-alpha.1227',
            'egg=django-thumborize-beta-1.0.0-alpha+beta',
            'egg=django-thumborize-beta-1.2.3----RC-SNAPSHOT.12.9.1--.12+788',  # noqa
            'egg=django-thumborize-beta-1.2.3----R-S.12.9.1--.12+meta',
            'egg=django-thumborize-beta-1.2.3----RC-SNAPSHOT.12.9.1--.12',
            'egg=django-thumborize-beta-1.0.0+0.build.1-rc.10000aaa-kk-0.1',  # noqa
            'egg=django-thumborize-beta-999999999999999999.99999999999999.9999999999999',  # noqa
            'egg=Proj1',
            'egg=Proj2-0.0.1',
            'egg=Proj3',
            'egg=Proj4-0.0.2',
            'egg=Proj5',
            'egg=Proj-0.0.3',
            'egg=Proj',
            'egg=Proj-0.0.4',
            'egg=Proj',
            'egg=foo-bar-1.2.4',
        ]
        for index, test in enumerate(tests):
            self.assertEqual(
                expected[index],
                re.sub(r'egg=([^&]+).*$', packaging.egg_fragment, test),
            )

    def test_parse_repo_url_requirements(self):
        result = packaging.parse_requirements([self.requirements])
        self.assertEqual(
            [
                'oslo.messaging>=1.0.0-rc',
                'django-thumborize',
                'django-thumborize-beta',
                'django-thumborize2-beta',
                'django-thumborize2-beta>=4.0.1',
                'django-thumborize2-beta>=1.0.0-alpha.beta.1',
                'django-thumborize2-beta>=1.0.0-alpha-a.b-c-somethinglong+build.1-aef.1-its-okay',  # noqa
                'django-thumborize2-beta>=2.0.0-rc.1+build.123',
                'Proj1',
                'Proj2>=0.0.1',
                'Proj3',
                'Proj4>=0.0.2',
                'Proj5',
                'Proj>=0.0.3',
                'Proj',
                'Proj>=0.0.4',
                'Proj',
                'foo-bar>=1.2.4',
                'pypi-proj1',
                'pypi-proj2',
            ],
            result,
        )

    def test_parse_repo_url_dependency_links(self):
        result = packaging.parse_dependency_links([self.requirements])
        self.assertEqual(
            [
                'git+git://git.pro-ject.org/oslo.messaging#egg=oslo.messaging-1.0.0-rc',  # noqa
                'git+git://git.pro-ject.org/django-thumborize#egg=django-thumborize',  # noqa
                'git+git://git.pro-ject.org/django-thumborize#egg=django-thumborize-beta',  # noqa
                'git+git://git.pro-ject.org/django-thumborize#egg=django-thumborize2-beta',  # noqa
                'git+git://git.pro-ject.org/django-thumborize#egg=django-thumborize2-beta-4.0.1',  # noqa
                'git+git://git.pro-ject.org/django-thumborize#egg=django-thumborize2-beta-1.0.0-alpha.beta.1',  # noqa
                'git+git://git.pro-ject.org/django-thumborize#egg=django-thumborize2-beta-1.0.0-alpha-a.b-c-somethinglong+build.1-aef.1-its-okay',  # noqa
                'git+git://git.pro-ject.org/django-thumborize#egg=django-thumborize2-beta-2.0.0-rc.1+build.123',  # noqa
                'git+git://git.project.org/Proj#egg=Proj1',
                'git+https://git.project.org/Proj#egg=Proj2-0.0.1',
                'git+ssh://git.project.org/Proj#egg=Proj3',
                'svn+svn://svn.project.org/svn/Proj#egg=Proj4-0.0.2',
                'svn+http://svn.project.org/svn/Proj/trunk@2019#egg=Proj5',
                'hg+http://hg.project.org/Proj@da39a3ee5e6b#egg=Proj-0.0.3',
                'hg+http://hg.project.org/Proj@2019#egg=Proj',
                'hg+http://hg.project.org/Proj@v1.0#egg=Proj-0.0.4',
                'hg+http://hg.project.org/Proj@special_feature#egg=Proj',
                'git://foo.com/zipball#egg=foo-bar-1.2.4',
            ],
            result,
        )



# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import
from __future__ import print_function

import pbr._compat.versions
from pbr.hooks import base
from pbr import options


class CommandsConfig(base.BaseConfig):

    section = 'global'

    def __init__(self, config):
        super(CommandsConfig, self).__init__(config)
        self.commands = self.config.get('commands', "")

    def save(self):
        self.config['commands'] = self.commands
        super(CommandsConfig, self).save()

    def add_command(self, command):
        self.commands = "%s\n%s" % (self.commands, command)

    def hook(self):
        self.add_command('pbr._compat.commands.LocalEggInfo')
        self.add_command('pbr._compat.commands.LocalSDist')
        self.add_command('pbr._compat.commands.LocalInstallScripts')
        self.add_command('pbr._compat.commands.LocalRPMVersion')
        self.add_command('pbr._compat.commands.LocalDebVersion')

        if pbr._compat.versions.setuptools_has_develop_command:
            self.add_command('pbr._compat.commands.LocalDevelop')

        use_egg = options.get_boolean_option(
            self.pbr_config, 'use-egg', 'PBR_USE_EGG'
        )
        # We always want non-egg install unless explicitly requested
        if 'manpages' in self.pbr_config or not use_egg:
            self.add_command('pbr._compat.commands.LocalInstall')
        else:
            self.add_command('pbr._compat.commands.InstallWithGit')



# Copyright 2010-2011 OpenStack Foundation
# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# Copyright (C) 2013 Association of Universities for Research in Astronomy
#                    (AURA)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided
#        with the distribution.
#
#     3. The name of AURA and its representatives may not be used to
#        endorse or promote products derived from this software without
#        specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS

"""Common utilities used in testing"""

from __future__ import absolute_import
from __future__ import print_function

import os
import shutil
import sys

import fixtures
import testresources
import testtools

from pbr import options


class BaseTestCase(testtools.TestCase, testresources.ResourcedTestCase):

    def setUp(self):
        super(BaseTestCase, self).setUp()
        test_timeout = os.environ.get('OS_TEST_TIMEOUT', 30)
        try:
            test_timeout = int(test_timeout)
        except ValueError:
            # If timeout value is invalid, fail hard.
            print(
                "OS_TEST_TIMEOUT set to invalid value"
                " defaulting to no timeout"
            )
            test_timeout = 0
        if test_timeout > 0:
            self.useFixture(fixtures.Timeout(test_timeout, gentle=True))

        if os.environ.get('OS_STDOUT_CAPTURE') in options.TRUE_VALUES:
            stdout = self.useFixture(fixtures.StringStream('stdout')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stdout', stdout))
        if os.environ.get('OS_STDERR_CAPTURE') in options.TRUE_VALUES:
            stderr = self.useFixture(fixtures.StringStream('stderr')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stderr', stderr))
        self.log_fixture = self.useFixture(fixtures.FakeLogger('pbr'))

        # Older git does not have config --local, so create a temporary home
        # directory to permit using git config --global without stepping on
        # developer configuration.
        self.useFixture(fixtures.TempHomeDir())
        self.useFixture(fixtures.NestedTempfile())
        self.useFixture(fixtures.FakeLogger())
        # TODO(lifeless) we should remove PBR_VERSION from the environment.
        # rather than setting it, because thats not representative - we need to
        # test non-preversioned codepaths too!
        self.useFixture(fixtures.EnvironmentVariable('PBR_VERSION', '0.0'))

        self.temp_dir = self.useFixture(fixtures.TempDir()).path
        self.package_dir = os.path.join(self.temp_dir, 'testpackage')
        shutil.copytree(
            os.path.join(os.path.dirname(__file__), 'testpackage'),
            self.package_dir,
        )
        self.addCleanup(os.chdir, os.getcwd())
        os.chdir(self.package_dir)
        self.addCleanup(self._discard_testpackage)
        # Tests can opt into non-PBR_VERSION by setting preversioned=False as
        # an attribute.
        if not getattr(self, 'preversioned', True):
            self.useFixture(fixtures.EnvironmentVariable('PBR_VERSION'))
            setup_cfg_path = os.path.join(self.package_dir, 'setup.cfg')
            with open(setup_cfg_path, 'rt') as cfg:
                content = cfg.read()
            content = content.replace(u'version = 0.1.dev', u'')
            with open(setup_cfg_path, 'wt') as cfg:
                cfg.write(content)

    def _discard_testpackage(self):
        # Remove pbr.testpackage from sys.modules so that it can be freshly
        # re-imported by the next test
        for k in list(sys.modules):
            if k == 'pbr_testpackage' or k.startswith('pbr_testpackage.'):
                del sys.modules[k]



# Copyright (c) 2013 New Dream Network, LLC (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright (C) 2013 Association of Universities for Research in Astronomy
#                    (AURA)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided
#        with the distribution.
#
#     3. The name of AURA and its representatives may not be used to
#        endorse or promote products derived from this software without
#        specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS

import os
import textwrap

import packaging.requirements

from pbr._compat.five import string_type
from pbr.tests import fixtures as pbr_fixtures
from pbr.tests.functional import base


# Taken from setuptools
#
# https://github.com/pypa/setuptools/blob/v44.1.1/pkg_resources/__init__.py#L2378-L2389
def yield_lines(strs):
    """Yield non-empty/non-comment lines of a string or sequence"""
    if isinstance(strs, string_type):
        for s in strs.splitlines():
            s = s.strip()
            # skip blank lines/comments
            if s and not s.startswith('#'):
                yield s
    else:
        for ss in strs:
            for s in yield_lines(ss):
                yield s


# Taken from setuptools
#
# https://github.com/pypa/setuptools/blob/v44.1.1/pkg_resources/__init__.py#L3189-L3212
def split_sections(s):
    """Split a string or iterable thereof into (section, content) pairs

    Each ``section`` is a stripped version of the section header ("[section]")
    and each ``content`` is a list of stripped lines excluding blank lines and
    comment-only lines.  If there are any such lines before the first section
    header, they're returned in a first ``section`` of ``None``.
    """
    section = None
    content = []
    for line in yield_lines(s):
        if line.startswith("["):
            if line.endswith("]"):
                if section or content:
                    yield section, content
                section = line[1:-1].strip()
                content = []
            else:
                raise ValueError("Invalid section heading", line)
        else:
            content.append(line)

    # wrap up last segment
    yield section, content


class TestRequirementParsing(base.BaseTestCase):

    def test_requirement_parsing(self):
        pkgs = {
            'test_reqparse': {
                'requirements.txt': textwrap.dedent(
                    """\
                        bar
                        quux<1.0; python_version=='2.6'
                        requests-aws>=0.1.4    # BSD License (3 clause)
                        Routes>=1.12.3,!=2.0,!=2.1;python_version=='2.7'
                        requests-kerberos>=0.6;python_version=='2.7' # MIT
                    """
                ),
                'setup.cfg': textwrap.dedent(
                    """\
                        [metadata]
                        name = test_reqparse

                        [extras]
                        test =
                            foo
                            baz>3.2 :python_version=='2.7' # MIT
                            bar>3.3 :python_version=='2.7' # MIT # Apache
                    """
                ),
            },
        }
        pkg_dirs = self.useFixture(pbr_fixtures.Packages(pkgs)).package_dirs
        pkg_dir = pkg_dirs['test_reqparse']
        # pkg_resources.split_sections uses None as the title of an
        # anonymous section instead of the empty string. Weird.
        expected_requirements = {
            None: ['bar', 'requests-aws>=0.1.4'],
            ":(python_version=='2.6')": ['quux<1.0'],
            ":(python_version=='2.7')": [
                'Routes!=2.0,!=2.1,>=1.12.3',
                'requests-kerberos>=0.6',
            ],
            'test': ['foo'],
            "test:(python_version=='2.7')": ['baz>3.2', 'bar>3.3'],
        }
        venv = self.useFixture(pbr_fixtures.Venv('reqParse'))
        bin_python = venv.python
        # Two things are tested by this
        # 1) pbr properly parses markers from requirements.txt and setup.cfg
        # 2) bdist_wheel causes pbr to not evaluate markers
        self._run_cmd(
            bin_python,
            ('setup.py', 'bdist_wheel'),
            allow_fail=False,
            cwd=pkg_dir,
        )
        egg_info = os.path.join(pkg_dir, 'test_reqparse.egg-info')

        requires_txt = os.path.join(egg_info, 'requires.txt')
        with open(requires_txt, 'rt') as requires:
            generated_requirements = dict(split_sections(requires))

        # NOTE(dhellmann): We have to spell out the comparison because
        # the rendering for version specifiers in a range is not
        # consistent across versions of setuptools.

        for section, expected in expected_requirements.items():
            # We wrap in str since we need packaging 22.0.0 or later to do
            # comparisons [1] and that doesn't support Python 2.7, 3.6
            #
            # https://github.com/pypa/packaging/commit/aebc072a06925cc0004b031e6b6f3028e5e2e686
            # https://pypi.org/project/packaging/22.0/
            exp_parsed = [
                str(packaging.requirements.Requirement(s)) for s in expected
            ]
            gen_parsed = [
                str(packaging.requirements.Requirement(s))
                for s in generated_requirements[section]
            ]
            self.assertEqual(exp_parsed, gen_parsed)



# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright (C) 2013 Association of Universities for Research in Astronomy
#                    (AURA)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided
#        with the distribution.
#
#     3. The name of AURA and its representatives may not be used to
#        endorse or promote products derived from this software without
#        specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS

from __future__ import absolute_import
from __future__ import print_function

import os
import sys

import fixtures

from pbr.tests.functional import base


class TestConsoleScripts(base.BaseTestCase):
    """Test generation of custom console scripts.

    We generate custom console scripts that do not rely on pkg_resources to
    handle imports. This is no longer a concern starting with pip 19.0, since
    pip handles generation of scripts for wheel and starting in 19.0, pip nows
    generate an intermediate wheel during installation.
    """

    cmd_names = ('pbr_test_cmd', 'pbr_test_cmd_with_class')

    def check_script_install(self, install_stdout):
        for cmd_name in self.cmd_names:
            install_txt = 'Installing %s script to %s' % (
                cmd_name,
                self.temp_dir,
            )
            self.assertIn(install_txt, install_stdout)

            cmd_filename = os.path.join(self.temp_dir, cmd_name)

            script_txt = open(cmd_filename, 'r').read()
            self.assertNotIn('pkg_resources', script_txt)

            stdout, _, return_code = self._run_cmd(cmd_filename)
            self.assertIn("PBR", stdout)

    def test_console_script_install(self):
        """Test that we install a non-pkg-resources console script."""

        if os.name == 'nt':
            self.skipTest('Windows support is passthrough')

        stdout, _, return_code = self.run_setup(
            'install_scripts', '--install-dir=%s' % self.temp_dir
        )

        self.useFixture(fixtures.EnvironmentVariable('PYTHONPATH', '.'))

        self.check_script_install(stdout)

    def test_console_script_develop(self):
        """Test that we develop a non-pkg-resources console script."""

        if sys.version_info < (3, 0):
            self.skipTest(
                'Fails with recent virtualenv due to '
                'https://github.com/pypa/virtualenv/issues/1638'
            )

        if os.name == 'nt':
            self.skipTest('Windows support is passthrough')

        # setuptools v80.0.0 switched to using pip for the 'develop' command,
        # which means easy_install is no longer invoked
        #
        # https://github.com/pypa/setuptools/commit/98e6b4cac625c6c13b718eeccea42d00d75f2577
        # https://setuptools.pypa.io/en/stable/history.html#v80-0-0
        if self.get_setuptools_version() >= (80, 0):
            self.skipTest('setuptools is too new')

        self.useFixture(
            fixtures.EnvironmentVariable('PYTHONPATH', ".:%s" % self.temp_dir)
        )

        stdout, _, return_code = self.run_setup(
            'develop', '--install-dir=%s' % self.temp_dir
        )

        self.check_script_install(stdout)



# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

import argparse
import functools
import sys


def application(env, start_response, data):
    sys.stderr.flush()  # Force the previous request log to be written.
    start_response('200 OK', [('Content-Type', 'text/html')])
    return [data.encode('utf-8')]


def main():
    parser = argparse.ArgumentParser(description='Return a string.')
    parser.add_argument(
        '--content', '-c', help='String returned', default='Hello World'
    )
    args = parser.parse_args()
    return functools.partial(application, data=args.content)


class WSGI(object):

    @classmethod
    def app(self):
        return functools.partial(application, data='Hello World')



# Copyright 2011 OpenStack Foundation
# Copyright 2012-2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Utilities with minimum-depends for use in setup.py
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import email
import email.errors
import os
import re
import sys
import warnings

from distutils import log

from pbr._compat.five import urlparse
import pbr._compat.packaging
from pbr import git
import pbr.pbr_json
from pbr import version

REQUIREMENTS_FILES = ('requirements.txt', 'tools/pip-requires')
PY_REQUIREMENTS_FILES = [
    x % sys.version_info[0]
    for x in ('requirements-py%d.txt', 'tools/pip-requires-py%d')
]
TEST_REQUIREMENTS_FILES = ('test-requirements.txt', 'tools/test-requires')


def get_requirements_files():
    files = os.environ.get("PBR_REQUIREMENTS_FILES")
    if files:
        return tuple(f.strip() for f in files.split(','))
    # Returns a list composed of:
    # - REQUIREMENTS_FILES with -py2 or -py3 in the name
    #   (e.g. requirements-py3.txt)
    # - REQUIREMENTS_FILES

    return PY_REQUIREMENTS_FILES + list(REQUIREMENTS_FILES)


def append_text_list(config, key, text_list):
    """Append a \n separated list to possibly existing value."""
    new_value = []
    current_value = config.get(key, "")
    if current_value:
        new_value.append(current_value)
    new_value.extend(text_list)
    config[key] = '\n'.join(new_value)


def _any_existing(file_list):
    return [f for f in file_list if os.path.exists(f)]


# Get requirements from the first file that exists
def get_reqs_from_files(requirements_files):
    existing = _any_existing(requirements_files)

    # TODO(stephenfin): Remove this in pbr 6.0+
    deprecated = [f for f in existing if f in PY_REQUIREMENTS_FILES]
    if deprecated:
        warnings.warn(
            'Support for \'-pyN\'-suffixed requirements files is '
            'removed in pbr 5.0 and these files are now ignored. '
            'Use environment markers instead. Conflicting files: '
            '%r' % deprecated,
            DeprecationWarning,
        )

    existing = [f for f in existing if f not in PY_REQUIREMENTS_FILES]
    for requirements_file in existing:
        with open(requirements_file, 'r') as fil:
            return fil.read().split('\n')

    return []


def egg_fragment(match):
    return re.sub(
        r'(?P<PackageName>[\w.-]+)-'
        r'(?P<GlobalVersion>'
        r'(?P<VersionTripple>'
        r'(?P<Major>0|[1-9][0-9]*)\.'
        r'(?P<Minor>0|[1-9][0-9]*)\.'
        r'(?P<Patch>0|[1-9][0-9]*)){1}'
        r'(?P<Tags>(?:\-'
        r'(?P<Prerelease>(?:(?=[0]{1}[0-9A-Za-z-]{0})(?:[0]{1})|'
        r'(?=[1-9]{1}[0-9]*[A-Za-z]{0})(?:[0-9]+)|'
        r'(?=[0-9]*[A-Za-z-]+[0-9A-Za-z-]*)(?:[0-9A-Za-z-]+)){1}'
        r'(?:\.(?=[0]{1}[0-9A-Za-z-]{0})(?:[0]{1})|'
        r'\.(?=[1-9]{1}[0-9]*[A-Za-z]{0})(?:[0-9]+)|'
        r'\.(?=[0-9]*[A-Za-z-]+[0-9A-Za-z-]*)'
        r'(?:[0-9A-Za-z-]+))*){1}){0,1}(?:\+'
        r'(?P<Meta>(?:[0-9A-Za-z-]+(?:\.[0-9A-Za-z-]+)*))){0,1}))',
        r'\g<PackageName>>=\g<GlobalVersion>',
        match.groups()[-1],
    )


def parse_requirements(requirements_files=None, strip_markers=False):
    if requirements_files is None:
        requirements_files = get_requirements_files()

    requirements = []
    for line in get_reqs_from_files(requirements_files):
        # Ignore comments
        if (not line.strip()) or line.startswith('#'):
            continue

        # Ignore index URL lines
        if re.match(
            r'^\s*(-i|--index-url|--extra-index-url|--find-links).*', line
        ):
            continue

        # Handle nested requirements files such as:
        # -r other-requirements.txt
        if line.startswith('-r'):
            req_file = line.partition(' ')[2]
            requirements += parse_requirements(
                [req_file], strip_markers=strip_markers
            )
            continue

        project_name = pbr._compat.packaging.extract_project_name(line)

        # For the requirements list, we need to inject only the portion
        # after egg= so that distutils knows the package it's looking for
        # such as:
        # -e git://github.com/openstack/nova/master#egg=nova
        # -e git://github.com/openstack/nova/master#egg=nova-1.2.3
        # -e git+https://foo.com/zipball#egg=bar&subdirectory=baz
        # http://github.com/openstack/nova/zipball/master#egg=nova
        # http://github.com/openstack/nova/zipball/master#egg=nova-1.2.3
        # git+https://foo.com/zipball#egg=bar&subdirectory=baz
        # git+[ssh]://github.com/openstack/nova/zipball/master#egg=nova-1.2.3
        # hg+[ssh]://github.com/openstack/nova/zipball/master#egg=nova-1.2.3
        # svn+[proto]://github.com/openstack/nova/zipball/master#egg=nova-1.2.3
        # -f lines are for index locations, and don't get used here
        if re.match(r'\s*-e\s+', line):
            extract = re.match(r'\s*-e\s+(.*)$', line)
            line = extract.group(1)
        egg = urlparse(line)
        if egg.scheme:
            line = re.sub(r'egg=([^&]+).*$', egg_fragment, egg.fragment)
        elif re.match(r'\s*-f\s+', line):
            line = None
            reason = 'Index Location'

        if line is not None:
            line = re.sub('#.*$', '', line)
            if strip_markers:
                semi_pos = line.find(';')
                if semi_pos < 0:
                    semi_pos = None
                line = line[:semi_pos]
            requirements.append(line)
        else:
            log.info('[pbr] Excluding %s: %s' % (project_name, reason))

    return requirements


def parse_dependency_links(requirements_files=None):
    if requirements_files is None:
        requirements_files = get_requirements_files()

    dependency_links = []
    # dependency_links inject alternate locations to find packages listed
    # in requirements
    for line in get_reqs_from_files(requirements_files):
        # skip comments and blank lines
        if re.match(r'(\s*#)|(\s*$)', line):
            continue
        # lines with -e or -f need the whole line, minus the flag
        if re.match(r'\s*-[ef]\s+', line):
            dependency_links.append(re.sub(r'\s*-[ef]\s+', '', line))
        # lines that are only urls can go in unmolested
        elif re.match(r'^\s*(https?|git(\+(https|ssh))?|svn|hg)\S*:', line):
            dependency_links.append(line)
    return dependency_links


def _get_increment_kwargs(git_dir, tag):
    """Calculate the sort of semver increment needed from git history.

    Every commit from HEAD to tag is consider for Sem-Ver metadata lines.
    See the pbr docs for their syntax.

    :return: a dict of kwargs for passing into SemanticVersion.increment.
    """
    result = {}
    if tag:
        version_spec = tag + "..HEAD"
    else:
        version_spec = "HEAD"

    # Get the raw body of the commit messages so that we don't have to
    # parse out any formatting whitespace and to avoid user settings on
    # git log output affecting out ability to have working sem ver headers.
    changelog = git._run_git_command(
        ['log', '--pretty=%B', version_spec], git_dir
    )
    symbols = set()
    header = 'sem-ver:'
    for line in changelog.split("\n"):
        line = line.lower().strip()
        if not line.lower().strip().startswith(header):
            continue
        new_symbols = line[len(header) :].strip().split(",")
        symbols.update([symbol.strip() for symbol in new_symbols])

    def _handle_symbol(symbol, symbols, impact):
        if symbol in symbols:
            result[impact] = True
            symbols.discard(symbol)

    _handle_symbol('bugfix', symbols, 'patch')
    _handle_symbol('feature', symbols, 'minor')
    _handle_symbol('deprecation', symbols, 'minor')
    _handle_symbol('api-break', symbols, 'major')
    for symbol in symbols:
        log.info('[pbr] Unknown Sem-Ver symbol %r' % symbol)
    # We don't want patch in the kwargs since it is not a keyword argument -
    # its the default minimum increment.
    result.pop('patch', None)
    return result


def _get_revno_and_last_tag(git_dir):
    """Return the commit data about the most recent tag.

    We use git-describe to find this out, but if there are no
    tags then we fall back to counting commits since the beginning
    of time.
    """
    changelog = git._iter_log_oneline(git_dir=git_dir)
    row_count = 0
    for row_count, (ignored, tag_set, ignored) in enumerate(changelog):
        version_tags = set()
        semver_to_tag = {}
        for tag in list(tag_set):
            try:
                semver = version.SemanticVersion.from_pip_string(tag)
                semver_to_tag[semver] = tag
                version_tags.add(semver)
            except Exception:
                pass

        if version_tags:
            return semver_to_tag[max(version_tags)], row_count

    return "", row_count


def _get_version_from_git_target(git_dir, target_version):
    """Calculate a version from a target version in git_dir.

    This is used for untagged versions only. A new version is calculated as
    necessary based on git metadata - distance to tags, current hash, contents
    of commit messages.

    :param git_dir: The git directory we're working from.
    :param target_version: If None, the last tagged version (or 0 if there are
        no tags yet) is incremented as needed to produce an appropriate target
        version following semver rules. Otherwise target_version is used as a
        constraint - if semver rules would result in a newer version then an
        exception is raised.
    :return: A semver version object.
    """
    tag, distance = _get_revno_and_last_tag(git_dir)
    last_semver = version.SemanticVersion.from_pip_string(tag or '0')
    if distance == 0:
        new_version = last_semver
    else:
        new_version = last_semver.increment(
            **_get_increment_kwargs(git_dir, tag)
        )
    if target_version is not None and new_version > target_version:
        raise ValueError(
            "git history requires a target version of %(new)s, but target "
            "version is %(target)s"
            % {'new': new_version, 'target': target_version}
        )
    if distance == 0:
        return last_semver
    new_dev = new_version.to_dev(distance)
    if target_version is not None:
        target_dev = target_version.to_dev(distance)
        if target_dev > new_dev:
            return target_dev
    return new_dev


def _get_version_from_git(pre_version=None):
    """Calculate a version string from git.

    If the revision is tagged, return that. Otherwise calculate a semantic
    version description of the tree.

    The number of revisions since the last tag is included in the dev counter
    in the version for untagged versions.

    :param pre_version: If supplied use this as the target version rather than
        inferring one from the last tag + commit messages.
    """
    git_dir = git._run_git_functions()
    if git_dir:
        try:
            tagged = git._run_git_command(
                ['describe', '--exact-match'], git_dir, throw_on_error=True
            ).replace('-', '.')
            target_version = version.SemanticVersion.from_pip_string(tagged)
        except Exception:
            if pre_version:
                # not released yet - use pre_version as the target
                target_version = version.SemanticVersion.from_pip_string(
                    pre_version
                )
            else:
                # not released yet - just calculate from git history
                target_version = None
        result = _get_version_from_git_target(git_dir, target_version)
        return result.release_string()
    # If we don't know the version, return an empty string so at least
    # the downstream users of the value always have the same type of
    # object to work with.
    try:
        return unicode()
    except NameError:
        return ''


def _get_version_from_pkg_metadata(package_name):
    """Get the version from package metadata if present.

    This looks for PKG-INFO if present (for sdists), and if not looks
    for METADATA (for wheels) and failing that will return None.
    """
    pkg_metadata_filenames = ['PKG-INFO', 'METADATA']
    pkg_metadata = {}
    for filename in pkg_metadata_filenames:
        try:
            with open(filename, 'r') as pkg_metadata_file:
                pkg_metadata = email.message_from_file(pkg_metadata_file)
        except (IOError, OSError, email.errors.MessageError):
            continue

    # Check to make sure we're in our own dir
    if pkg_metadata.get('Name', None) != package_name:
        return None
    return pkg_metadata.get('Version', None)


def get_version(package_name, pre_version=None):
    """Get the version of the project.

    First, try getting it from PKG-INFO or METADATA, if it exists. If it does,
    that means we're in a distribution tarball or that install has happened.
    Otherwise, if there is no PKG-INFO or METADATA file, pull the version
    from git.

    We do not support setup.py version sanity in git archive tarballs, nor do
    we support packagers directly sucking our git repo into theirs. We expect
    that a source tarball be made from our git repo - or that if someone wants
    to make a source tarball from a fork of our repo with additional tags in it
    that they understand and desire the results of doing that.

    :param pre_version: The version field from setup.cfg - if set then this
        version will be the next release.
    """
    version = os.environ.get(
        "PBR_VERSION", os.environ.get("OSLO_PACKAGE_VERSION", None)
    )
    if version:
        return version
    version = _get_version_from_pkg_metadata(package_name)
    if version:
        return version
    version = _get_version_from_git(pre_version)
    # Handle http://bugs.python.org/issue11638
    # version will either be an empty unicode string or a valid
    # unicode version string, but either way it's unicode and needs to
    # be encoded.
    if sys.version_info[0] == 2:
        version = version.encode('utf-8')
    if version:
        return version
    raise Exception(
        "Versioning for this project requires either an sdist "
        "tarball, or access to an upstream git repository. "
        "It's also possible that there is a mismatch between "
        "the package name in setup.cfg and the argument given "
        "to pbr.version.VersionInfo. Project name {name} was "
        "given, but was not able to be found.".format(name=package_name)
    )


# This is added because pbr uses pbr to install itself. That means that
# any changes to the egg info writer entrypoints must be forward and
# backward compatible. This maintains the pbr.packaging.write_pbr_json
# path.
write_pbr_json = pbr.pbr_json.write_pbr_json



# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright (C) 2013 Association of Universities for Research in Astronomy
#                    (AURA)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided
#        with the distribution.
#
#     3. The name of AURA and its representatives may not be used to
#        endorse or promote products derived from this software without
#        specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS

from __future__ import absolute_import
from __future__ import print_function

import contextlib
import os
import shutil
import stat
import subprocess
import sys

from pbr._compat.five import ConfigParser


@contextlib.contextmanager
def open_config(filename):
    cfg = ConfigParser()
    cfg.read(filename)
    yield cfg
    with open(filename, 'w') as fp:
        cfg.write(fp)


def rmtree(path):
    """shutil.rmtree() with error handler.

    Handle 'access denied' from trying to delete read-only files.
    """

    def onexc(func, path, exc_info):
        if not os.access(path, os.W_OK):
            os.chmod(path, stat.S_IWUSR)
            func(path)
        else:
            raise

    if sys.version_info >= (3, 12):
        return shutil.rmtree(path, onexc=onexc)
    else:
        return shutil.rmtree(path, onerror=onexc)


def run_cmd(args, cwd):
    """Run the command args in cwd.

    :param args: The command to run e.g. ['git', 'status']
    :param cwd: The directory to run the command in.
    :param env: The environment variables to set. If unset, fallback to the
        default of inheriting those from the current process.
    :return: ((stdout, stderr), returncode)
    """
    env = os.environ.copy()
    env['PYTHONWARNINGS'] = 'ignore'

    print('Running %s' % ' '.join(args))
    p = subprocess.Popen(
        args,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=cwd,
        env=env,
    )
    streams = tuple(s.decode('latin1').strip() for s in p.communicate())
    print('STDOUT:')
    print(streams[0])
    print('STDERR:')
    print(streams[1])
    return (streams) + (p.returncode,)


def config_git():
    run_cmd(
        ['git', 'config', '--global', 'user.email', 'example@example.com'],
        None,
    )
    run_cmd(
        ['git', 'config', '--global', 'user.name', 'OpenStack Developer'], None
    )
    run_cmd(
        [
            'git',
            'config',
            '--global',
            'user.signingkey',
            'example@example.com',
        ],
        None,
    )
